{*******************************************************************
   SYSTEM.PAS    Runtime kernel for Pascal

   Copyright (c) 1998 by Mill Hill & Canterbury Corporation, Ltd.
********************************************************************}


UNIT SYSTEM;


INTERFACE

{ the interface itself is part of the compiler... }


IMPLEMENTATION

USES
  java_lang_RuntimeException,
  java_io_InputStreamReader,
  java_io_OutputStreamWriter,
  java_lang_Exception;


CONST
  IsLongReal      = SizeOf( REAL ) = SizeOf( DOUBLE );

TYPE
  OPENINDEX       = 0..HIGH(LONGINT)-1;
  BYTEARRAY       = ARRAY [OPENINDEX] OF BYTE;
  PBYTEARRAY      = ^BYTEARRAY;
  EInvalidOp      = CLASS( RuntimeException ) END;
  ENotImplemented = CLASS( RuntimeException ) END;
  EInOutError     = CLASS( RuntimeException ) END;
  EInvalidCast    = CLASS( RuntimeException ) END;
  EConvertError   = CLASS( RuntimeException ) END;


VAR
  InOutRes         : INTEGER;
  args             : POINTER;
  stdtext          : ^TEXT;
  stdin            : InputStreamReader;
  stdout           : OutputStreamWriter;
  vENotImplemented : ENotImplemented;
  termCH           : INTEGER;



PROCEDURE OrBytes( Dest,Source1,Source2:POINTER; Size:LONGWORD );
VAR
  i,j       : LONGINT;
  val       : PBYTEARRAY;
  val1      : PBYTEARRAY;
  val2      : PBYTEARRAY;
BEGIN
  { Dest := Source1 OR Source2 for non-overlapping operands }
  val := Dest;
  val1 := Source1;
  val2 := Source2;
  i := 0;
  j := Size;
  WHILE i < j DO BEGIN
    val^[i] := val1^[i] OR val2^[i];
    INC( i );
  END;
END;


PROCEDURE XorBytes( Dest,Source1,Source2:POINTER; Size:LONGWORD );
VAR
  i,j       : LONGINT;
  val       : PBYTEARRAY;
  val1      : PBYTEARRAY;
  val2      : PBYTEARRAY;
BEGIN
  { Dest := Source1 XOR Source2 for non-overlapping operands }
  val := Dest;
  val1 := Source1;
  val2 := Source2;
  i := 0;
  j := Size;
  WHILE i < j DO BEGIN
    val^[i] := val1^[i] XOR val2^[i];
    INC( i );
  END;
END;


PROCEDURE AndBytes( Dest,Source1,Source2:POINTER; Size:LONGWORD );
VAR
  i,j       : LONGINT;
  val       : PBYTEARRAY;
  val1      : PBYTEARRAY;
  val2      : PBYTEARRAY;
BEGIN
  { Dest := Source1 AND Source2  for non-overlapping operands }
  val := Dest;
  val1 := Source1;
  val2 := Source2;
  i := 0;
  j := Size;
  WHILE i < j DO BEGIN
    val^[i] := val1^[i] AND val2^[i];
    INC( i );
  END;
END;


PROCEDURE MemSet( Dest:POINTER; Val:BYTE; Size:LONGWORD );
VAR
  b   : PBYTEARRAY;
  i,j : LONGINT;
BEGIN
  { Set multiple memory bytes to a given value }
  i := 0;
  j := Size;
  b := Dest;
  WHILE i < j DO BEGIN
    b^[i] := Val;
    INC( i );
  END;
END;


FUNCTION MemCmp( Dest,Source: POINTER; Size:LONGWORD ) : LONGINT;
VAR
  val1 : PBYTEARRAY;
  val2 : PBYTEARRAY;
  i,j  : LONGINT;
BEGIN
  { compare two memory regions and return result as follows:  }
  { -1: Dest < Source   0: Dest == Source   1: Dest > Source  }
  val1 := Dest;
  val2 := Source;
  i := 0;
  j := Size;
  WHILE TRUE DO BEGIN
    IF i >= j THEN BEGIN
      BREAK;
    END;
    IF val1^[i] <> val2^[i] THEN BEGIN
      BREAK;
    END;
    INC( i );
  END;
  IF i >= j THEN BEGIN
    Result := 0;
  END
  ELSE IF val1^[i] < val2^[i] THEN BEGIN
    Result := -1;
  END
  ELSE BEGIN
    Result := 1;
  END;
END;


PROCEDURE MemCpy( Dest,Source: POINTER; Size:LONGWORD );
VAR
  val1 : PBYTEARRAY;
  val2 : PBYTEARRAY;
  i,j  : LONGINT;
BEGIN
  { copy muliple bytes from one memory region to another. }
  val1 := Dest;
  val2 := Source;
  i := 0;
  j := Size;
  WHILE i < j DO BEGIN
    val1^[i] := val2^[i];
    INC( i );
  END;
END;


FUNCTION SubsetOf( Set1, Set2 : POINTER; Size : LONGWORD ) : BOOLEAN;
BEGIN
  { runtime function for implementing set1 <= set2 relation }
  INLINE
  (
    byte[] t = new byte[Size];
    AndBytes( t, Set1, Set2, Size );
    Result = (MemCmp( t, Set1, Size ) == 0);
  );
END;



FUNCTION AddSet( Set1, Set2 : POINTER; Size : LONGWORD ) : POINTER;
BEGIN
  { runtime function for implementing set union }
  INLINE
  (
    byte[] t = new byte[Size];
    OrBytes( t, Set1, Set2, Size );
    Result = t;
  );
END;


FUNCTION SubSet( Set1, Set2 : POINTER; Size : LONGWORD ) : POINTER;
BEGIN
  { runtime function for implementing set difference }
  INLINE
  (
    byte[] t = new byte[Size];
    OrBytes( t, Set1, Set2, Size );
    XorBytes( t, t, Set2, Size );
    Result = t;
  );
END;



FUNCTION MulSet( Set1, Set2 : POINTER; Size : LONGWORD ) : POINTER;
BEGIN
  { runtime function for implementing set intersection }
  INLINE
  (
    byte[] t = new byte[Size];
    AndBytes( t, Set1, Set2, Size );
    Result = t;
  );
END;


FUNCTION DivSet( Set1, Set2 : POINTER; Size : LONGWORD ) : POINTER;
BEGIN
  { runtime function for implementing set symmetric difference }
  INLINE
  (
    byte[] t = new byte[Size];
    XorBytes( t, Set1, Set2, Size );
    Result = t;
  );
END;


FUNCTION StrCmp( CONST s1, s2 : ARRAY OF CHAR ) : LONGINT;
BEGIN
  { runtime function for comparing 2 strings, returning: }
  { 1: s1 > s2    0: s1 == s2    -1: s1 < s2             }
  INLINE
  (
    int j = s1.length;
    int k = s2.length;
    int i = 0;
    Result = 0;
    if (j < k)
      Result = -1;
    else if (j > k)
      Result = 1;
    else {
      while (i<j) {
        if (s1[i]<s2[i])
          Result = -1;
        else if (s1[i]>s2[i])
          Result = 1;
        if (Result != 0)
          break;
        else
          i++;
        };
      };
  );
END;


FUNCTION MUL_1( x,y : BYTE ) : BYTE;
BEGIN
  { runtime function for unsigned 8-bit integer multiplication }
  INLINE
  (
    int i,j,k;
    i = (int)x & 0x000000FF;
    j = (int)y & 0x000000FF;
    k = i * j;
    Result = (byte)k;
  );
END;


FUNCTION MUL_2( x,y : WORD ) : WORD;
BEGIN
  { runtime function for unsigned 16-bit integer multiplication }
  INLINE
  (
    int i,j,k;

    i = (int)x & 0x0000FFFF;
    j = (int)y & 0x0000FFFF;
    k = i * j;
    Result = (short)k;
  );
END;


FUNCTION MUL_4( x,y : LONGWORD ) : LONGWORD;
BEGIN
  { runtime function for unsigned 32-bit integer multiplication }
  INLINE
  (
    long i,j,k;

    i = (long)x & 0xFFFFFFFFL;
    j = (long)y & 0xFFFFFFFFL;
    k = i * j;
    Result = (int)k;
  );
END;



FUNCTION DIV_1( x,y : BYTE ) : BYTE;
BEGIN
  { runtime function for unsigned 8-bit integer division }
  INLINE
  (
    int i,j,k;

    i = (int)x & 0x000000FF;
    j = (int)y & 0x000000FF;
    k = i / j;
    Result = (byte)k;
  );
END;


FUNCTION DIV_2( x,y : WORD ) : WORD;
BEGIN
  { runtime function for unsigned 16-bit integer division }
  INLINE
  (
    int i,j,k;

    i = (int)x & 0x0000FFFF;
    j = (int)y & 0x0000FFFF;
    k = i / j;
    Result = (short)k;
  );
END;


FUNCTION DIV_4( x,y : LONGWORD ) : LONGWORD;
BEGIN
  { runtime function for unsigned 32-bit integer division }
  INLINE
  (
    long i,j,k;

    i = (long)x & 0xFFFFFFFFL;
    j = (long)y & 0xFFFFFFFFL;
    k = i / j;
    Result = (int)k;
  );
END;


FUNCTION MOD_1( x,y : BYTE ) : BYTE;
BEGIN
  { runtime function for unsigned 8-bit integer division remainder }
  INLINE
  (
    int i,j,k;

    i = (int)x & 0x000000FF;
    j = (int)y & 0x000000FF;
    k = i % j;
    Result = (byte)k;
  );
END;


FUNCTION MOD_2( x,y : WORD ) : WORD;
BEGIN
  { runtime function for unsigned 16-bit integer division remainder }
  INLINE
  (
    int i,j,k;

    i = (int)x & 0x0000FFFF;
    j = (int)y & 0x0000FFFF;
    k = i % j;
    Result = (short)k;
  );
END;


FUNCTION MOD_4( x,y : LONGWORD ) : LONGWORD;
BEGIN
  { runtime function for unsigned 32-bit integer division remainder }
  INLINE
  (
    long i,j,k;

    i = (long)x & 0xFFFFFFFFL;
    j = (long)y & 0xFFFFFFFFL;
    k = i % j;
    Result = (int)k;
  );
END;


FUNCTION CMP_1( x,y : BYTE ) : LONGINT;
BEGIN
  { runtime function for unsigned 8-bit integer compare }
  INLINE
  (
    int i,j;

    i = (int)x & 0x000000FF;
    j = (int)y & 0x000000FF;
    if (i < j)
    {
      Result = -1;
    }
    else if (i == j)
    {
      Result = 0;
    }
    else
    {
      Result = 1;
    }
  );
END;


FUNCTION CMP_2( x,y : WORD ) : LONGINT;
BEGIN
  { runtime function for unsigned 16-bit integer compare }
  INLINE
  (
    int i,j;

    i = (int)x & 0x0000FFFF;
    j = (int)y & 0x0000FFFF;
    if (i < j)
    {
      Result = -1;
    }
    else if (i == j)
    {
      Result = 0;
    }
    else
    {
      Result = 1;
    }
  );
END;


FUNCTION CMP_4( x,y : LONGWORD ) : LONGINT;
BEGIN
  { runtime function for unsigned 32-bit integer compare }
  INLINE
  (
    long i,j;

    i = (long)x & 0xFFFFFFFFL;
    j = (long)y & 0xFFFFFFFFL;
    if (i < j)
    {
      Result = -1;
    }
    else if (i == j)
    {
      Result = 0;
    }
    else
    {
      Result = 1;
    }
  );
END;


TYPE
  BITARRAY = ARRAY [0..31] OF BYTE;


PROCEDURE SetBitRange
(
  Dest    : POINTER;
  FromBit : LONGWORD;
  ToBit   : LONGWORD;
  MaxBit  : LONGWORD
);
VAR
  setval  : ^BITARRAY;
  i       : LONGWORD;
BEGIN
  setval := Dest;
  i := SIZEOF( setval^ ) * 8 - 1;
  IF i < MaxBit THEN BEGIN
    MaxBit := i;
  END;
  IF ToBit > MaxBit THEN BEGIN
    ToBit := MaxBit;
  END;
  i := FromBit;
  WHILE i <= ToBit DO BEGIN
    setval^[i DIV 8] := setval^[i DIV 8] OR BYTE(1 SHL (i MOD 8));
    i := i + 1;
  END;
END;



PROCEDURE DelBitRange
(
  Dest    : POINTER;
  FromBit : LONGWORD;
  ToBit   : LONGWORD;
  MaxBit  : LONGWORD
);
VAR
  setval  : ^BITARRAY;
  i       : LONGWORD;
BEGIN
  setval := Dest;
  i := SIZEOF( setval^ ) * 8 - 1;
  IF i < MaxBit THEN BEGIN
    MaxBit := i;
  END;
  IF ToBit > MaxBit THEN BEGIN
    ToBit := MaxBit;
  END;
  i := FromBit;
  WHILE i <= ToBit DO BEGIN
    setval^[i DIV 8] := setval^[i DIV 8] AND NOT BYTE(1 SHL (i MOD 8));
    i := i + 1;
  END;
END;



FUNCTION TestBit( BSet:POINTER; BitPos:LONGWORD; BitMax:LONGWORD ):BOOLEAN;
VAR
  setval  : ^BITARRAY;
  val     : BYTE;
BEGIN
  { Test a bit of a bitset; Return TRUE if set else return FALSE }
  IF BitPos > BitMax THEN BEGIN
    TestBit := FALSE;
  END
  ELSE BEGIN
    setval := BSet;
    val := setval^[ BitPos SHR 3 ];
    TestBit := ((val AND (1 SHL (BYTE( BitPos ) AND 7))) > 0);
  END;
END;




FUNCTION set1( s : PSET; i : LONGINT ) : PSET;
BEGIN
  { runtime function for including a bit in a given set }
  INLINE
  (
    s[i / 8] |= (byte)(1 << (i % 8));
    Result = s;
  );
END;



FUNCTION set2( s : PSET; i : LONGINT; j : LONGINT ) : PSET;
BEGIN
  { runtime function for including bits in a given set }
  INLINE
  (
    int k = s.length * 8;
    while ((i <= j) && (i < k))
    {
      s[i / 8] |= (byte)(1 << (i % 8));
      i++;
    }
    Result = s;
  )
END;


PROCEDURE VAL( s : STRING; VAR n; VAR code : INTEGER );
BEGIN
  INLINE
  (
    try
    {
      InOutRes = 0;
      code[0] = 0;
      if (n instanceof byte[])
      {
        byte[] m = (byte[])n;
        java.lang.String str = s.toString();
        int i = str.indexOf( (int)'$' );
        int j = str.indexOf( (int)'-' );
        int k = (i >= 0) ? 16 : 10;
        m[0] = (byte)java.lang.Integer.parseInt( str.substring(i+1), k );
        if (j >= 0)
        {
          m[0] = (byte)(-m[0]);
        }
      }
      else if (n instanceof short[])
      {
        short[] m = (short[])n;
        java.lang.String str = s.toString();
        int i = str.indexOf( (int)'$' );
        int j = str.indexOf( (int)'-' );
        int k = (i >= 0) ? 16 : 10;
        m[0] = (short)java.lang.Integer.parseInt( str.substring(i+1), k );
        if (j >= 0)
        {
          m[0] = (short)(-m[0]);
        }
      }
      else if (n instanceof int[])
      {
        int[] m = (int[])n;
        java.lang.String str = s.toString();
        int i = str.indexOf( (int)'$' );
        int j = str.indexOf( (int)'-' );
        int k = (i >= 0) ? 16 : 10;
        m[0] = (int)java.lang.Integer.parseInt( str.substring(i+1), k );
        if (j >= 0)
        {
          m[0] = -m[0];
        }
      }
      else if (n instanceof float[])
      {
        float[] m = (float[])n;
        java.lang.String str = s.toString();
        int j = ((str.length() > 0) && (str.charAt(0) == '-')) ? 1 : 0;
        m[0] = java.lang.Float.valueOf( str.substring(j) ).floatValue();
        if (j > 0)
        {
          m[0] = -m[0];
        }
      }
      else if (n instanceof double[])
      {
        double[] m = (double[])n;
        java.lang.String str = s.toString();
        int j = ((str.length() > 0) && (str.charAt(0) == '-')) ? 1 : 0;
        m[0] = java.lang.Double.valueOf( str.substring(j) ).doubleValue();
        if (j > 0)
        {
          m[0] = -m[0];
        }
      }
      else
      {
        code[0] = 1;
        InOutRes = 1;
      }
    }
    catch (java.lang.NumberFormatException e)
    {
      InOutRes = 1;
      code[0] = 1;
    }
    catch (java.lang.NullPointerException e)
    {
      InOutRes = 1;
      code[0] = 1;
    }
  );
END;


PROCEDURE WriteInteger( VAR f : TEXT; val : LONGINT; size : INTEGER );
BEGIN
  INLINE
  (
    java.lang.String s = java.lang.Integer.toString( val );
    int i = size - s.length();
    while (i > 0)
    {
      i -= 1;
      s = " " + s;
    }
    WriteString( f, s.toCharArray() );
  );
END;


PROCEDURE WriteCardinal( VAR f : TEXT; val : LONGWORD; size : INTEGER );
BEGIN
  INLINE
  (
    java.lang.String s = java.lang.Long.toString( (long)val & 0xFFFFFFFFL );
    int i = size - s.length();
    while (i > 0)
    {
      i -= 1;
      s = " " + s;
    }
    WriteString( f, s.toCharArray() );
  );
END;


PROCEDURE WriteEOL( VAR f : TEXT );
BEGIN
  INLINE
  (
    java.lang.String  eol;
    try
    {
      eol = java.lang.System.getProperty( "line.separator" );
    }
    catch (java.lang.SecurityException e)
    {
      eol = null;
    }
    if (eol == null)
    {
      eol = "\n";
    }
    WriteString( f, eol.toCharArray() );
  );
END;



PROCEDURE WriteSingle( VAR f : TEXT; val : SINGLE; size1,size2: INTEGER );
VAR
  s : STRING;
BEGIN
  (* Write float value as decimal number:            *)
  (*   [-]digit{digit}.digit{digit}                  *)
  (* or as a floating point notation:                *)
  (*  [  |-]<digit>.<decimals>E[+|-]<exponent>       *)
  RealToStr( val, s, size1, size2, FALSE );
  INLINE
  (
    WriteString( f, (new java.lang.String( s )).toCharArray() );
  );
END;


PROCEDURE WriteDouble( VAR f : TEXT; val : DOUBLE; size1,size2: INTEGER );
VAR
  s : STRING;
BEGIN
  (* Write float value as decimal number:            *)
  (*   [-]digit{digit}.digit{digit}                  *)
  (* or as a floating point notation:                *)
  (*  [  |-]<digit>.<decimals>E[+|-]<exponent>       *)
  RealToStr( val, s, size1, size2, TRUE );
  INLINE
  (
    WriteString( f, (new java.lang.String( s )).toCharArray() );
  );
END;


PROCEDURE WriteString( VAR f : TEXT; CONST val : ARRAY OF CHAR );
BEGIN
  INLINE
  (
    if (f == null)
    {
      f = stdtext;
    }
  );
  IF (f.Handle <> NIL) OR (f.FileName <> NIL) THEN BEGIN
    IF InOutRes <> 0 THEN BEGIN
      EXIT;
    END;
  END;
  INLINE
  (
    boolean ok;
    int i = 0;
    int j = val.length;
    while ((i<j) && (val[i] != (char)0))
    {
      i += 1;
    }
    if ((f.Handle == null) && (f.FileName == null))
    {
      try
      {
        stdout.write( val, 0, i );
        stdout.flush();
      }
      catch (java.io.IOException e)
      { }
      ok = true;
    }
    else if (f.Handle != null)
    {
      try
      {
        ((java.io.RandomAccessFile)f.Handle).writeBytes( new java.lang.String( val, 0, i ) );
        ok = true;
      }
      catch (java.io.IOException e)
      {
        ok = false;
      }
    }
    else
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;


PROCEDURE WriteBool( VAR f : TEXT; val : BOOLEAN );
BEGIN
  IF val THEN BEGIN
    WriteString( f, 'TRUE' );
  END
  ELSE BEGIN
    WriteString( f, 'FALSE' );
  END;
END;


PROCEDURE WriteChar( VAR f : TEXT; val : CHAR );
BEGIN
  INLINE
  (
    if (f == null)
    {
      f = stdtext;
    }
  );
  IF (f.Handle <> NIL) OR (f.FileName <> NIL) THEN BEGIN
    IF InOutRes <> 0 THEN BEGIN
      EXIT;
    END;
  END;
  INLINE
  (
    boolean ok;
    if ((f.Handle == null) && (f.FileName == null))
    {
      try
      {
        stdout.write( (int)val );
        stdout.flush();
      }
      catch (java.io.IOException e)
      { }
      ok = true;
    }
    else if (f.Handle != null)
    {
      try
      {
        ((java.io.RandomAccessFile)f.Handle).writeByte( (int)val );
        ok = true;
      }
      catch (java.io.IOException e)
      {
        ok = false;
      }
    }
    else
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;



PROCEDURE ReadShortInt( VAR f : TEXT; VAR val : SHORTINT );
VAR
  i : LONGINT;
BEGIN
  ReadLongInt( f, i );
  IF (i < -128) OR (i > 127) THEN BEGIN
    InOutRes := 1;
  END;
  val  := SHORTINT(i);
END;



PROCEDURE ReadSmallInt( VAR f : TEXT; VAR val : SMALLINT );
VAR
  i : LONGINT;
BEGIN
  ReadLongInt( f, i );
  IF (i < -32768) OR (i > 32767) THEN BEGIN
    InOutRes := 1;
  END;
  val  := SMALLINT(i);
END;



PROCEDURE SkipBlanks( VAR f : TEXT; VAR ch : CHAR );
CONST
  TAB = CHR( 9 );
  CR  = CHR( 13 );
  BS  = CHR( 8 );
VAR
  IsStdIn : BOOLEAN;
BEGIN
  IsStdIn := (f.Handle = NIL) AND (f.FileName = NIL);
  REPEAT
    IF IsStdIn AND (termCH <> -1) THEN BEGIN
      ch := CHR(termCH);
      termCH := -1;
    END
    ELSE BEGIN
      ReadChar( f, ch );
    END;
  UNTIL (ch <> ' ') AND (ch <> TAB) AND (ch <> CR) AND (ch <> BS);
END;




PROCEDURE ReadLongInt( VAR f : TEXT; VAR val : LONGINT );
VAR
  ch         : CHAR;
  i          : HUGEINT;
  IsNegative : BOOLEAN;
  digits     : INTEGER;
BEGIN
  INLINE
  (
    if (f == null)
    {
      f = stdtext;
    }
  );
  IF (f.Handle <> NIL) OR (f.FileName <> NIL) THEN BEGIN
    IF InOutRes <> 0 THEN BEGIN
      EXIT;
    END;
  END;
  SkipBlanks( f, ch );
  IF (ch = '-') THEN BEGIN
    ReadChar( f, ch );
    IsNegative := TRUE;
  END
  ELSE BEGIN
    IsNegative := FALSE;
  END;
  INLINE
  (
    i = 0L;
  );
  digits := 0;
  WHILE ('0' <= ch) AND (ch <= '9') DO BEGIN
    INLINE
    (
      i = i * 10L + (long)(ch-'0');
    );
    INC( digits );
    ReadChar( f, ch );
  END;
  termCH := ORD(ch);
  INLINE
  (
    boolean ok = true;
    if (digits == 0)
    {
      ok = false;
    }
    if (IsNegative)
    {
      if ((i < 0L) || (i > (long)java.lang.Integer.MAX_VALUE+1L))
      {
        ok = false;
      }
      i = -i;
    }
    else if (i < (long)java.lang.Integer.MIN_VALUE)
    {
      ok = false;
    }
    else if (i > (long)java.lang.Integer.MAX_VALUE)
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
    val[0] = (int)i;
  );
END;


PROCEDURE ReadByte( VAR f : TEXT; VAR val : BYTE );
VAR
  i : LONGWORD;
BEGIN
  ReadLongWord( f, i );
  IF (i > 255) THEN BEGIN
    InOutRes := 1;
  END;
  val  := BYTE(i);
END;


PROCEDURE ReadWord( VAR f : TEXT; VAR val : WORD );
VAR
  i : LONGWORD;
BEGIN
  ReadLongWord( f, i );
  IF (i > 65535) THEN BEGIN
    InOutRes := 1;
  END;
  val  := WORD(i);
END;


PROCEDURE ReadLongWord( VAR f : TEXT; VAR val : LONGWORD );
VAR
  ch      : CHAR;
  i       : HUGEINT;
  digits  : INTEGER;
BEGIN
  INLINE
  (
    if (f == null)
    {
      f = stdtext;
    }
  );
  IF (f.Handle <> NIL) OR (f.FileName <> NIL) THEN BEGIN
    IF InOutRes <> 0 THEN BEGIN
      EXIT;
    END;
  END;
  SkipBlanks( f, ch );
  INLINE
  (
    i = 0L;
  );
  digits := 0;
  WHILE ('0' <= ch) AND (ch <= '9') DO BEGIN
    INLINE
    (
      i = i * 10L + (long)ch - (long)'0';
    );
    INC( digits );
    ReadChar( f, ch );
  END;
  termCH := ORD(ch);
  INLINE
  (
    boolean ok = true;
    if ((i & 0xFFFFFFFF00000000L) != 0L)
    {
      ok = false;
    }
    if (digits == 0)
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
    val[0] = (int)i;
  );
END;


PROCEDURE ReadEOL( VAR f : TEXT );
VAR
  IsStdIn : BOOLEAN;
  ch      : CHAR;
BEGIN
  INLINE
  (
    if (f == null)
    {
      f = stdtext;
    }
  );
  IF (f.Handle <> NIL) OR (f.FileName <> NIL) THEN BEGIN
    IF InOutRes <> 0 THEN BEGIN
      EXIT;
    END;
  END;
  IsStdIn := (f.Handle = NIL) AND (f.FileName = NIL);
  REPEAT
    IF IsStdIn AND (termCH <> -1) THEN BEGIN
      ch := CHR(termCH);
      termCH := -1;
    END
    ELSE BEGIN
      ReadChar( f, ch );
    END;
  UNTIL (ch = CHR(10)) OR (ch = CHR(0));
  IF (ch = CHR(0)) THEN BEGIN
    InOutRes := 1;
  END;
END;



PROCEDURE ReadSubString( VAR f : TEXT; VAR s : ARRAY OF CHAR );
CONST
  TAB = CHR( 9 );
  CR  = CHR( 13 );
  BS  = CHR( 8 );
  EOL = CHR( 10 );
VAR
  ch      : CHAR;
  IsStdIn : BOOLEAN;
  i,j     : LONGINT;
  ok      : BOOLEAN;
BEGIN
  s[0] := CHR(0);
  INLINE
  (
    if (f == null)
    {
      f = stdtext;
    }
  );
  IF (f.Handle <> NIL) OR (f.FileName <> NIL) THEN BEGIN
    IF InOutRes <> 0 THEN BEGIN
      EXIT;
    END;
  END;
  SkipBlanks( f, ch );
  i := 0;
  j := HIGH(s);
  WHILE (ch<>' ') AND (ch<>TAB) AND (ch<>CR) AND
        (ch<>BS)  AND (ch<>EOL) AND (ch<>CHR(0)) DO BEGIN
    IF i < j THEN BEGIN
      s[i] := ch;
    END;
    INC(i);
    ReadChar( f, ch );
  END;
  IsStdIn := (f.Handle = NIL) AND (f.FileName = NIL);
  IF IsStdIn THEN BEGIN
    termCH := ORD(ch);
  END;
  ok := TRUE;
  IF i >= j THEN BEGIN
    ok := FALSE;
    s[j] := CHR(0);
  END
  ELSE BEGIN
    s[i] := CHR(0);
  END;
  IF ch = CHR(0) THEN BEGIN
    ok := FALSE;
  END;
  IF i = 0 THEN BEGIN
    ok := FALSE;
  END;
  IF NOT ok THEN BEGIN
    InOutRes := 1;
  END;
END;




FUNCTION LONGFLOAT( x : LONGINT ) : DOUBLE;
BEGIN
  INLINE
  (
    Result = (double)x;
  );
END;


CONST
  SGLDIG           = 6;        (* max # of significand dec. digits   *)
  DBLDIG           = 15;       (* max # of significand dec. digits   *)
  DIGMAX           = DBLDIG*2; (* max # of digits in string          *)
                               (* ( *2 is a good fudge factor)       *)
  DIGPREC          = DBLDIG+2; (* max # of significant digits        *)
                               (* (+2 for fractional part after min  *)
                               (*     significant digits)            *)


CONST
  negtab  : ARRAY [0..9] OF DOUBLE =
  (1.E-256, 1.E-128, 1.E-64, 1.E-32, 1.E-16, 1.E-8, 1.E-4, 1.E-2, 1.E-1, 1.0);
  postab  : ARRAY [0..9] OF DOUBLE =
  (1.E+256, 1.E+128, 1.E+64, 1.E+32, 1.E+16, 1.E+8, 1.E+4, 1.E+2, 1.E+1);

TYPE
  DIGITSTRING      = ARRAY [0..DIGMAX-1 +1 +1] OF CHAR;
                     (* DIGMAX digits in a string   *)
                     (* +1 for end of string        *)
                     (* +1 in case rounding adds    *)



FUNCTION IsLongNAN( r : DOUBLE ) : BOOLEAN;
BEGIN
  INLINE
  (
    Result = java.lang.Double.isNaN(r);
  );
END;



FUNCTION IsLongINFINITY( r : DOUBLE ) : BOOLEAN;
BEGIN
  INLINE
  (
    Result = java.lang.Double.isInfinite( r );
  );
END;


(*@IF FALSE *)

FUNCTION IsShortNAN( r : SINGLE ) : BOOLEAN;
BEGIN
  INLINE
  (
    Result = java.lang.Float.isNaN(r);
  );
END;



FUNCTION IsShortINFINITY( r : SINGLE ) : BOOLEAN;
BEGIN
  INLINE
  (
    Result = java.lang.Float.isInfinite( r );
  );
END;



PROCEDURE StrToLongReal
( s        : ARRAY OF CHAR;
  VAR r    : DOUBLE;
  VAR done : BOOLEAN
);
VAR
  p       : LONGWORD;
  dval    : DOUBLE;
  exp     : SMALLINT;
  msdec   : LONGWORD;
  lsdec   : LONGWORD;
  msscale : LONGWORD;
  dot     : BOOLEAN;
  sign    : BOOLEAN;
  subject : BOOLEAN;
  pow     : SMALLINT;
  ch      : CHAR;
  sexp    : BOOLEAN;
  e       : SMALLINT;
  u       : WORD;
CONST
  MAXEXP  = 1024;
  HT      = CHR( 9  );
  LF      = CHR( 10 );
  VT      = CHR( 11 );
  FF      = CHR( 12 );
  CR      = CHR( 13 );
BEGIN
  s[ HIGH(s) ] := CHR(0);
  p := 0;
  WHILE (s[p]=' ') OR (s[p]=HT) OR (s[p]=LF) OR (s[p]=VT) OR (s[p]=FF) OR (s[p]=CR) DO BEGIN
    INC( p );
  END;
  sign := FALSE;              (* indicating + *)
  IF s[p] = '-' THEN BEGIN
    sign := TRUE;
    INC( p );
  END
  ELSE IF s[p] = '+' THEN BEGIN
    INC( p );
  END;
  subject := FALSE;
  dval := 0.0;
  dot := FALSE;               (* if decimal point has been seen *)
  exp := 0;
  msdec := 0;
  lsdec := 0;
  msscale := 1;
  WHILE TRUE DO BEGIN
    ch := s[p];
    WHILE (('0' <= ch) AND (ch <= '9')) DO BEGIN
      subject := TRUE; (* must have at least 1 digit *)
      IF msdec < (LONGWORD(HIGH(LONGINT))-10) DIV 10 THEN BEGIN
        msdec := msdec * 10 + LONGWORD( ORD(ch) - ORD('0') );
      END
      ELSE IF msscale < (LONGWORD(HIGH(LONGINT))-10) DIV 10 THEN BEGIN
        lsdec := lsdec * 10 + LONGWORD( ORD(ch) - ORD('0') );
        msscale := msscale * 10;
      END
      ELSE BEGIN
        INC( exp );
      END;
      IF dot THEN BEGIN
        DEC( exp );
      END;
      INC( p );
      ch := s[p];
    END;
    IF (ch = '.') AND (NOT dot) THEN BEGIN
      INC( p );
      dot := TRUE;
    END
    ELSE BEGIN
      BREAK;
    END;
  END;
  IF NOT subject THEN BEGIN   (* if error (no digits seen) *)
    r := 0.0;                 (* return 0.0 *)
    done := FALSE; EXIT;
  END;
  IF (s[p] = 'E') OR (s[p] = 'e') THEN BEGIN
    sexp := FALSE;
    INC( p );
    IF s[p] = '-' THEN BEGIN
      sexp := TRUE;
      INC ( p );
    END
    ELSE IF s[p] = '+' THEN BEGIN
      INC( p );
    END;
    subject := FALSE;
    e := 0;
    ch := s[p];
    WHILE ('0' <= ch) AND (ch <= '9') DO BEGIN
      IF e < MAXEXP*2 THEN BEGIN (* prevent integer overflow *)
        e := e * 10 + SMALLINT( ORD( ch ) - ORD( '0' ) );
      END;
      INC( p );
      subject := TRUE;
      ch := s[p];
    END;
    IF sexp THEN BEGIN
      exp := exp - e;
    END
    ELSE BEGIN
      exp := exp + e;
    END;
    IF NOT subject THEN BEGIN  (* if no digits in exponent *)
      r := 0.0;                (* return 0.0 *)
      done := FALSE; EXIT;
    END;
  END;
  dval := LONGFLOAT( LONGINT(msdec) );
  IF msscale <> 1 THEN BEGIN  (* if stuff was accumulated in lsdec *)
    dval := dval * LONGFLOAT( LONGINT( msscale ) );
    dval := dval + LONGFLOAT( LONGINT( lsdec ) );
  END;
  IF dval <> 0.0 THEN BEGIN
    u := 0;
    pow := 256;
    WHILE exp > 0 DO BEGIN
      WHILE exp >= pow DO BEGIN
        dval := dval * postab[ u ];
        exp := exp - pow;
      END;
      pow := pow DIV 2;
      INC( u );
    END;
    WHILE exp < 0 DO BEGIN
      WHILE exp <= -pow DO BEGIN
        dval := dval * negtab[ u ];
        IF dval = 0.0 THEN BEGIN
          r := dval;
          done := FALSE; EXIT;
        END;
        exp := exp + pow;
      END;
      pow := pow DIV 2;
      INC( u );
    END;
  END;
  IF sign THEN BEGIN
    r := -dval;
  END
  ELSE BEGIN
    r := dval;
  END;
  done := NOT IsLongNAN( dval ) AND NOT IsLongINFINITY( dval );
END;



PROCEDURE StrToShortReal
( s        : ARRAY OF CHAR;
  VAR r    : SINGLE;
  VAR done : BOOLEAN
);
VAR
  dval    : DOUBLE;
  ok      : BOOLEAN;
BEGIN
  StrToLongReal( s, dval, ok );
  r := SINGLE( dval );
  done := ok AND NOT IsShortNAN( r ) AND NOT IsShortINFINITY( r );
END;
(*@ENDIF *)

FUNCTION StringsLength
( VAR Str          : ARRAY OF CHAR
)                  : CARDINAL;
VAR
  i                : CARDINAL;
BEGIN
  i := 0;
  WHILE (i <= HIGH( Str )) AND (Str[ i ] <> CHR(0)) DO BEGIN
    INC( i );
  END;
  Result := i;
END;



PROCEDURE StrToLongReal
( s        : ARRAY OF CHAR;
  VAR r    : DOUBLE;
  VAR done : BOOLEAN
);
BEGIN
  INLINE
  (
    int i = StringsLength( s );
    if (i > 0)
    {
      java.lang.String str = new java.lang.String(s,0,i);
      try
      {
        r[0] = java.lang.Double.valueOf( str ).doubleValue();
        done[0] = true;
      }
      catch (java.lang.NumberFormatException e)
      {
        r[0] = 0.0;
        done[0] = false;
      }
    }
    else
    {
      r[0] = 0.0;
      done[0] = false;
    }
  );
END;



PROCEDURE StrToShortReal
( s        : ARRAY OF CHAR;
  VAR r    : SINGLE;
  VAR done : BOOLEAN
);
BEGIN
  INLINE
  (
    int i = StringsLength( s );
    if (i > 0)
    {
      java.lang.String str = new java.lang.String(s,0,i);
      try
      {
        r[0] = java.lang.Float.valueOf( str ).floatValue();
        done[0] = true;
      }
      catch (java.lang.NumberFormatException e)
      {
        r[0] = (float)0.0;
        done[0] = false;
      }
    }
    else
    {
      r[0] = (float)0.0;
      done[0] = false;
    }
  );
END;



PROCEDURE ReadSingle( VAR f : TEXT; VAR val : SINGLE );
VAR
  Buffer : ARRAY [0..40] OF CHAR;
  Done   : BOOLEAN;
BEGIN
  INLINE
  (
    if (f == null)
    {
      f = stdtext;
    }
  );
  IF (f.Handle <> NIL) OR (f.FileName <> NIL) THEN BEGIN
    IF InOutRes <> 0 THEN BEGIN
      EXIT;
    END;
  END;
  Buffer[0] := CHR(0);
  ReadSubString( f, Buffer );
  Done := InOutRes = 0;
  IF Done THEN BEGIN
    StrToShortReal( Buffer, val, Done );
  END;
  IF NOT Done THEN BEGIN
    InOutRes := 1;
  END;
END;


PROCEDURE ReadDouble( VAR f : TEXT; VAR val : DOUBLE );
VAR
  Buffer : ARRAY [0..40] OF CHAR;
  Done   : BOOLEAN;
BEGIN
  INLINE
  (
    if (f == null)
    {
      f = stdtext;
    }
  );
  IF (f.Handle <> NIL) OR (f.FileName <> NIL) THEN BEGIN
    IF InOutRes <> 0 THEN BEGIN
      EXIT;
    END;
  END;
  Buffer[0] := CHR(0);
  ReadSubString( f, Buffer );
  Done := InOutRes = 0;
  IF Done THEN BEGIN
    StrToLongReal( Buffer, val, Done );
  END;
  IF NOT Done THEN BEGIN
    InOutRes := 1;
  END;
END;


PROCEDURE ReadString( VAR f : TEXT; VAR val );
BEGIN
  INLINE
  (
    if (f == null)
    {
      f = stdtext;
    }
  );
  IF (f.Handle <> NIL) OR (f.FileName <> NIL) THEN BEGIN
    IF InOutRes <> 0 THEN BEGIN
      EXIT;
    END;
  END;
  INLINE
  (
    java.lang.String s;
    boolean ok;
    int i;
    if ((f.Handle == null) && (f.FileName == null))
    {
      char[] ch = new char[1];
      java.lang.StringBuffer b = new java.lang.StringBuffer();
      while (true)
      {
        if (termCH != -1)
        {
          ch[0] = (char)termCH;
          termCH = -1;
        }
        else
        {
          ReadChar( f, ch );
        }
        if (ch[0] == '\n')
        {
          termCH = ((int)ch[0]) & 0xFFFF;
          break;
        }
        if (ch[0] != '\r')
        {
          b.append( ch[0] );
        }
      }
      s = b.toString();
      ok = true;
    }
    else if (f.Handle != null)
    {
      try
      {
        s = ((java.io.RandomAccessFile)f.Handle).readLine();
        long pos = ((java.io.RandomAccessFile)f.Handle).getFilePointer();
        if (pos >= 0L)
        {
          ((java.io.RandomAccessFile)f.Handle).seek(pos-1);
          char c = (char)((java.io.RandomAccessFile)f.Handle).read();
          if (c == '\n')
          {
            ((java.io.RandomAccessFile)f.Handle).seek(pos-1);
          }
        }
        ok = true;
      }
      catch (java.io.IOException e)
      {
        ok = false;
        s = "";
      }
    }
    else
    {
      ok = false;
      s = "";
    }
    i = s.length();
    if (i > 0)
    {
      if (s.charAt(i-1) == '\r')
      {
        i -= 1;
        s = s.substring( 0, i );
      }
    }
    if (val instanceof char[])
    {
      if (i > ((char[])val).length)
      {
        i = ((char[])val).length;
        s = s.substring( 0, i );
      }
    }
    else /* val instanceof java.lang.StringBuffer */
    {
//    if (i > ((java.lang.StringBuffer)val).capacity())
//    {
//      i = ((java.lang.StringBuffer)val).length();
//      s = s.substring( 0, i );
//    }
    }

    if (!ok)
    {
      InOutRes = 1;
    }

    if (val instanceof char[])
    {
      s.getChars( 0, i, (char[])val, 0 );
    }
    else
    {
      java.lang.StringBuffer str = (java.lang.StringBuffer)val;
      str.setLength( 0 );
      str.append( s );
    }
  );
END;


VAR
  lookAhead : INTEGER;

PROCEDURE ReadChar( VAR f : TEXT; VAR val : CHAR );
BEGIN
  INLINE
  (
    if (f == null)
    {
      f = stdtext;
    }
  );
  IF (f.Handle <> NIL) OR (f.FileName <> NIL) THEN BEGIN
    IF InOutRes <> 0 THEN BEGIN
      EXIT;
    END;
  END;
  INLINE
  (
    boolean ok;
    if ((f.Handle == null) && (f.FileName == null))
    {
      try
      {
        if (lookAhead==-1)
        {
          val[0] = (char)stdin.read();
        }
        else
        {
          val[0] = (char)lookAhead;
          lookAhead = -1;
        }
      }
      catch (java.io.IOException e)
      {
        val[0] = (char)0;
      }
      ok = true;
    }
    else if (f.Handle != null)
    {
      try
      {
        val[0] = (char)((java.io.RandomAccessFile)f.Handle).read();
        ok = true;
      }
      catch (java.io.IOException e)
      {
        ok = false;
        val[0] = (char)0;
      }
    }
    else
    {
      ok = false;
      val[0] = (char)0;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;



PROCEDURE WriteFileBool( VAR f : FILE; val : BOOLEAN );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok;
    try
    {
      if (f.Handle != null)
      {
        ((java.io.RandomAccessFile)f.Handle).writeBoolean( val );
        ok = true;
      }
      else
      {
        ok = false;
      }
    }
    catch (java.io.IOException e)
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;


PROCEDURE WriteFileChar( VAR f : FILE; val : CHAR );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok;
    try
    {
      if (f.Handle != null)
      {
        ((java.io.RandomAccessFile)f.Handle).writeChar( (int)val );
        ok = true;
      }
      else
      {
        ok = false;
      }
    }
    catch (java.io.IOException e)
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;


PROCEDURE WriteFileShortint( VAR f : FILE; val : SHORTINT );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok;
    try
    {
      if (f.Handle != null)
      {
        ((java.io.RandomAccessFile)f.Handle).writeByte( (int)val );
        ok = true;
      }
      else
      {
        ok = false;
      }
    }
    catch (java.io.IOException e)
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;


PROCEDURE WriteFileSmallint( VAR f : FILE; val : SMALLINT );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok;
    try
    {
      if (f.Handle != null)
      {
        ((java.io.RandomAccessFile)f.Handle).writeShort( (int)val );
        ok = true;
      }
      else
      {
        ok = false;
      }
    }
    catch (java.io.IOException e)
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;


PROCEDURE WriteFileLongint( VAR f : FILE; val : LONGINT );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok;
    try
    {
      if (f.Handle != null)
      {
        ((java.io.RandomAccessFile)f.Handle).writeInt( val );
        ok = true;
      }
      else
      {
        ok = false;
      }
    }
    catch (java.io.IOException e)
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;


PROCEDURE WriteFileByte( VAR f : FILE; val : BYTE );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok;
    try
    {
      if (f.Handle != null)
      {
        ((java.io.RandomAccessFile)f.Handle).writeByte( (int)val );
        ok = true;
      }
      else
      {
        ok = false;
      }
    }
    catch (java.io.IOException e)
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;


PROCEDURE WriteFileWord( VAR f : FILE; val : WORD );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok;
    try
    {
      if (f.Handle != null)
      {
        ((java.io.RandomAccessFile)f.Handle).writeShort( (int)val );
        ok = true;
      }
      else
      {
        ok = false;
      }
    }
    catch (java.io.IOException e)
    {
      ok = false;
    }  // <-------------- obsolete goto next generated
    if (!ok)
    {
      InOutRes = 1;
    }
  );

END;


PROCEDURE WriteFileLongword( VAR f : FILE; val : LONGWORD );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok;
    try
    {
      if (f.Handle != null)
      {
        ((java.io.RandomAccessFile)f.Handle).writeInt( val );
        ok = true;
      }
      else
      {
        ok = false;
      }
    }
    catch (java.io.IOException e)
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;


PROCEDURE WriteFileSingle( VAR f : FILE; val : SINGLE );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok;
    try
    {
      if (f.Handle != null)
      {
        ((java.io.RandomAccessFile)f.Handle).writeFloat( val );
        ok = true;
      }
      else
      {
        ok = false;
      }
    }
    catch (java.io.IOException e)
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;


PROCEDURE WriteFileDouble( VAR f : FILE; val : DOUBLE );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok;
    try
    {
      if (f.Handle != null)
      {
        ((java.io.RandomAccessFile)f.Handle).writeDouble( val );
        ok = true;
      }
      else
      {
        ok = false;
      }
    }
    catch (java.io.IOException e)
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;



PROCEDURE WriteFileObject( VAR f : FILE; VAR val );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    java.lang.Class c = val.getClass();
    int i = 1;
    if (c.isArray())
    {
      if (c.getComponentType() == java.lang.Character.TYPE)
      {
        i = ((char[])val).length;
      }
      else if (c.getComponentType() == java.lang.Boolean.TYPE)
      {
        i = ((boolean[])val).length;
      }
      else if (c.getComponentType() == java.lang.Byte.TYPE)
      {
        i = ((byte[])val).length;
      }
      else if (c.getComponentType() == java.lang.Short.TYPE)
      {
        i = ((short[])val).length;
      }
      else if (c.getComponentType() == java.lang.Integer.TYPE)
      {
        i = ((int[])val).length;
      }
      else if (c.getComponentType() == java.lang.Long.TYPE)
      {
        i = ((long[])val).length;
      }
      else if (c.getComponentType() == java.lang.Float.TYPE)
      {
        i = ((float[])val).length;
      }
      else if (c.getComponentType() == java.lang.Double.TYPE)
      {
        i = ((double[])val).length;
      }
      else if (!c.getComponentType().isPrimitive())
      {
        i = ((java.lang.Object[])val).length;
      }
      else
      {
        InOutRes = 1;
      }
    }
    else if (c.isPrimitive())
    {
      InOutRes = 1;
    }
    if (InOutRes == 0)
    {
      int[] res = new int[1];
      BlockWrite( f, val, i, res );
      res = null;
    }
  );
END;


PROCEDURE ReadFileBool( VAR f : FILE; VAR val : BOOLEAN );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok;
    try
    {
      if (f.Handle != null)
      {
        val[0] = ((java.io.RandomAccessFile)f.Handle).readBoolean();
        ok = true;
      }
      else
      {
        ok = false;
      }
    }
    catch (java.io.IOException e)
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;


PROCEDURE ReadFileChar( VAR f : FILE; VAR val : CHAR );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok;
    try
    {
      if (f.Handle != null)
      {
        val[0] = ((java.io.RandomAccessFile)f.Handle).readChar();
        ok = true;
      }
      else
      {
        ok = false;
      }
    }
    catch (java.io.IOException e)
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;


PROCEDURE ReadFileShortint( VAR f : FILE; VAR val : SHORTINT );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok;
    try
    {
      if (f.Handle != null)
      {
        val[0] = ((java.io.RandomAccessFile)f.Handle).readByte();
        ok = true;
      }
      else
      {
        ok = false;
      }
    }
    catch (java.io.IOException e)
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;


PROCEDURE ReadFileSmallint( VAR f : FILE; VAR val : SMALLINT );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok;
    try
    {
      if (f.Handle != null)
      {
        val[0] = ((java.io.RandomAccessFile)f.Handle).readShort();
        ok = true;
      }
      else
      {
        ok = false;
      }
    }
    catch (java.io.IOException e)
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;


PROCEDURE ReadFileLongint( VAR f : FILE; VAR val : LONGINT );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok;
    try
    {
      if (f.Handle != null)
      {
        val[0] = ((java.io.RandomAccessFile)f.Handle).readInt();
        ok = true;
      }
      else
      {
        ok = false;
      }
    }
    catch (java.io.IOException e)
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;


PROCEDURE ReadFileByte( VAR f : FILE; VAR val : BYTE );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok;
    try
    {
      if (f.Handle != null)
      {
        val[0] = (byte)((java.io.RandomAccessFile)f.Handle).readUnsignedByte();
        ok = true;
      }
      else
      {
        ok = false;
      }
    }
    catch (java.io.IOException e)
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;


PROCEDURE ReadFileWord( VAR f : FILE; VAR val : WORD );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok;
    try
    {
      if (f.Handle != null)
      {
        val[0] = (short)((java.io.RandomAccessFile)f.Handle).readUnsignedShort();
        ok = true;
      }
      else
      {
        ok = false;
      }
    }
    catch (java.io.IOException e)
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;


PROCEDURE ReadFileLongword( VAR f : FILE; VAR val : LONGWORD );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok;
    try
    {
      if (f.Handle != null)
      {
        val[0] = ((java.io.RandomAccessFile)f.Handle).readInt();
        ok = true;
      }
      else
      {
        ok = false;
      }
    }
    catch (java.io.IOException e)
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;


PROCEDURE ReadFileSingle( VAR f : FILE; VAR val : SINGLE );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok;
    try
    {
      if (f.Handle != null)
      {
        val[0] = ((java.io.RandomAccessFile)f.Handle).readFloat();
        ok = true;
      }
      else
      {
        ok = false;
      }
    }
    catch (java.io.IOException e)
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;


PROCEDURE ReadFileDouble( VAR f : FILE; VAR val : DOUBLE );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok;
    try
    {
      if (f.Handle != null)
      {
        val[0] = ((java.io.RandomAccessFile)f.Handle).readDouble();
        ok = true;
      }
      else
      {
        ok = false;
      }
    }
    catch (java.io.IOException e)
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;



PROCEDURE CopyObject( VAR dest; VAR origin; VAR done : BOOLEAN );
BEGIN
  INLINE
  (
    java.lang.Class destClass = dest.getClass();
    java.lang.Class originClass = origin.getClass();
    if (originClass != destClass)
    {
      done[0] = false;
    }
    else if (originClass.isArray())
    {
      if (originClass.getComponentType().isPrimitive())
      {
        try
        {
          done[0] = true;
          java.lang.Class c = originClass.getComponentType();
          if (c == java.lang.Character.TYPE)
          {
            java.lang.System.arraycopy
            ( origin, 0, dest, 0, ((char[])origin).length );
          }
          else if (c == java.lang.Boolean.TYPE)
          {
            java.lang.System.arraycopy
            ( origin, 0, dest, 0, ((boolean[])origin).length );
          }
          else if (c == java.lang.Byte.TYPE)
          {
            java.lang.System.arraycopy
            ( origin, 0, dest, 0, ((byte[])origin).length );
          }
          else if (c == java.lang.Short.TYPE)
          {
            java.lang.System.arraycopy
            ( origin, 0, dest, 0, ((short[])origin).length );
          }
          else if (c == java.lang.Integer.TYPE)
          {
            java.lang.System.arraycopy
            ( origin, 0, dest, 0, ((int[])origin).length );
          }
          else if (c == java.lang.Long.TYPE)
          {
            java.lang.System.arraycopy
            ( origin, 0, dest, 0, ((long[])origin).length );
          }
          else if (c == java.lang.Float.TYPE)
          {
            java.lang.System.arraycopy
            ( origin, 0, dest, 0, ((float[])origin).length );
          }
          else if (c == java.lang.Double.TYPE)
          {
            java.lang.System.arraycopy
            ( origin, 0, dest, 0, ((double[])origin).length );
          }
          else
          {
            done[0] = false;
          }
        }
        catch (java.lang.Exception e)
        {
          done[0] = false;
        }
      }
      else
      {
        try
        {
          done[0] = true;
          int j = ((java.lang.Object[])origin).length;
          int i;
          for (i=0; ((i<j) && (done[0])); i++)
          {
            CopyObject
            ( ((java.lang.Object[])dest)[i],
              ((java.lang.Object[])origin)[i],
              done
            );
          }
        }
        catch (java.lang.Exception e)
        {
          done[0] = false;
        }
      }
    }
    else if (!originClass.isPrimitive())
    {
      java.lang.reflect.Method m;
      java.lang.Class[] paramType = new java.lang.Class[1];
      paramType[0] = destClass;
      try
      {
        m = destClass.getDeclaredMethod( "_copy", paramType );
      }
      catch (java.lang.Exception e)
      {
        m = null;
      }
      if (m != null)
      {
        try
        {
          java.lang.Object[] args = new java.lang.Object[1];
          args[0] = origin;
          m.invoke( dest, args );
          done[0] = true;
        }
        catch (java.lang.Exception e)
        {
          done[0] = false;
        }
      }
      else
      {
        done[0] = false;
      }
    }
    else
    {
      done[0] = false;
    }
  );
END;




PROCEDURE ReadFileObject( VAR f : FILE; VAR val );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    java.lang.Class c = val.getClass();
    int i = 1;
    if (c.isArray())
    {
      if (c.getComponentType() == java.lang.Character.TYPE)
      {
        i = ((char[])val).length;
      }
      else if (c.getComponentType() == java.lang.Boolean.TYPE)
      {
        i = ((boolean[])val).length;
      }
      else if (c.getComponentType() == java.lang.Byte.TYPE)
      {
        i = ((byte[])val).length;
      }
      else if (c.getComponentType() == java.lang.Short.TYPE)
      {
        i = ((short[])val).length;
      }
      else if (c.getComponentType() == java.lang.Integer.TYPE)
      {
        i = ((int[])val).length;
      }
      else if (c.getComponentType() == java.lang.Long.TYPE)
      {
        i = ((long[])val).length;
      }
      else if (c.getComponentType() == java.lang.Float.TYPE)
      {
        i = ((float[])val).length;
      }
      else if (c.getComponentType() == java.lang.Double.TYPE)
      {
        i = ((double[])val).length;
      }
      else if (!c.getComponentType().isPrimitive())
      {
        i = ((java.lang.Object[])val).length;
      }
      else
      {
        InOutRes = 1;
      }
    }
    else if (c.isPrimitive())
    {
      InOutRes = 1;
    }
    if (InOutRes == 0)
    {
      int[] res = new int[1];
      BlockRead( f, val, i, res );
      res = null;
    }
  );
END;



PROCEDURE SetLength( VAR S : STRING; Len : INTEGER );
BEGIN
  INLINE
  (
    S.setLength( Len );
  );
END;




FUNCTION forName( id : ARRAY OF CHAR ) : POINTER;
BEGIN
  { Runtime function for finding and loading a class }
  { Only used for passing the result to IsEqualProcs }
  INLINE
  (
    try
    {
      Result = java.lang.Class.forName( new java.lang.String( id ) );
    }
    catch (java.lang.ClassNotFoundException e)
    {
      Result = null;
    }
  );
END;



FUNCTION IsEqualProcs( p1, p2 : POINTER ) : BOOLEAN;
BEGIN
(*
    Runtime function for testing equality of 2 procedure pointers:

      p1 = instance of procedure-pointer class or
           instance of procedure-pointer meta-class

      p2 = instance of procedure-pointer class or
           instance of procedure-pointer meta-class

    Procedure types are implemented as classes.
    A specific procedure address is implemented as an instance
    of an extended procedure type class. This extended class
    always has a name like this:

      <proc-type-id>_<mod-id>_ADR_<proc-id>

    Procedure pointer equality is tested by comparing the
    name parts starting with
    "_ADR".

    NOTE: This runtime function is obsolete for compiler
          versions >= 3.0
*)
  INLINE
  (
    java.lang.String s1;
    java.lang.String s2;
    int i1;
    int i2;

    while (true)
    {
      if ( (p1 instanceof java.lang.Class) &&
           (p2 instanceof java.lang.Class) ||
           (p1 instanceof java.lang.Object) &&
           (p2 instanceof java.lang.Object)
         )
      {
        if (p1 == p2)
        {
          Result = true;
          break;
        }
      }

      if (p1 instanceof java.lang.Class)
      {
        s1 = ((java.lang.Class)p1).getName();
      }
      else if (p1 instanceof java.lang.Object)
      {
        s1 = p1.getClass().getName();
      }
      else
      {
        Result = (p1 == p2);
        break;
      }

      if (p2 instanceof java.lang.Class)
      {
        s2 = ((java.lang.Class)p2).getName();
      }
      else if (p2 instanceof java.lang.Object)
      {
        s2 = p2.getClass().getName();
      }
      else
      {
        Result = (p1 ==p2);
        break;
      }

      if (s1.equals( s2 ))
      {
        Result = true;
        break;
      }

      i1 = s1.indexOf( "_ADR" );
      if (i1 < 0)
      {
        Result = false;
        break;
      }
      i2 = s2.indexOf( "_ADR" );
      if (i2 < 0)
      {
        Result = false;
        break;
      }
      s1 = s1.substring( i1 );
      s2 = s2.substring( i2 );
      Result = (s1.equals( s2 ));
      break;
    }
  );
END;



FUNCTION Frac( r : DOUBLE ) : DOUBLE;
BEGIN
  INLINE
  (
    boolean negative = false;
    if (r < 0.0)
    {
      negative = true;
      r = -r;
    }
    Result = r - java.lang.Math.floor(r);
    if (negative)
    {
      Result = -Result;
    }
  );
END;


FUNCTION Int( r : DOUBLE ) : DOUBLE;
BEGIN
  INLINE
  (
    boolean negative = false;
    if (r < 0.0)
    {
      negative = true;
      r = -r;
    }
    Result = java.lang.Math.floor(r);
    if (negative)
    {
      Result = -Result;
    }
  );
END;



FUNCTION Round( r : DOUBLE ) : LONGINT;
BEGIN
  INLINE
  (
    r = java.lang.Math.rint(r);
    if ( (r < (double)java.lang.Integer.MIN_VALUE) ||
         (r > (double)java.lang.Integer.MAX_VALUE)
       )
    {
      throw new SYSTEM_EInvalidOp();
    }
  );
  Result := Trunc(r);
END;


FUNCTION Trunc( r : DOUBLE ) : LONGINT;
BEGIN
  INLINE
  (
    r = Int(r);
    if ( (r < (double)java.lang.Integer.MIN_VALUE) ||
         (r > (double)java.lang.Integer.MAX_VALUE)
       )
    {
      throw new SYSTEM_EInvalidOp();
    }
    Result = (int)java.lang.Math.round(r);
  );
END;


FUNCTION Sqr( r : DOUBLE ) : DOUBLE;
BEGIN
  Sqr := r * r;
END;


FUNCTION Sqrt( r : DOUBLE ) : DOUBLE;
BEGIN
  INLINE
  (
    Result = java.lang.Math.sqrt(r);
  );
END;


FUNCTION Exp( r : DOUBLE ) : DOUBLE;
BEGIN
  INLINE
  (
    Result = java.lang.Math.exp(r);
  )
END;


FUNCTION Ln( r : DOUBLE ) : DOUBLE;
BEGIN
  INLINE
  (
    Result = java.lang.Math.log(r);
  );
END;


FUNCTION Sin( r : DOUBLE ) : DOUBLE;
BEGIN
  INLINE
  (
    Result = java.lang.Math.sin(r);
  );
END;


FUNCTION Cos( r : DOUBLE ) : DOUBLE;
BEGIN
  INLINE
  (
    Result = java.lang.Math.cos(r);
  );
END;


FUNCTION Tan( r : DOUBLE ) : DOUBLE;
BEGIN
  INLINE
  (
    Result = java.lang.Math.tan(r);
  );
END;



FUNCTION ArcSin( r : DOUBLE ) : DOUBLE;
BEGIN
  INLINE
  (
    Result = java.lang.Math.asin(r);
  );
END;



FUNCTION Pi : DOUBLE;
BEGIN
  INLINE
  (
    Result = java.lang.Math.PI;
  );
END;



FUNCTION ArcCos( r : DOUBLE ) : DOUBLE;
BEGIN
  INLINE
  (
    Result = java.lang.Math.acos(r);
  );
END;


FUNCTION ArcTan( r : DOUBLE ) : DOUBLE;
BEGIN
  INLINE
  (
    Result = java.lang.Math.atan(r);
  );
END;



PROCEDURE Append( VAR f : TEXT );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    if ((f.FileName == null) || (f.FileName.length() == 0))
    {
      f.Handle = null;
      InOutRes = 1;
    }
    else
    {
      java.io.File handle = new java.io.File( new java.lang.String( f.FileName ) );
      try
      {
        if ((handle.exists()) || (handle.isFile()))
        {
          java.io.RandomAccessFile r = new java.io.RandomAccessFile( handle, "rw" );
          r.close();
          r = new java.io.RandomAccessFile( handle, "rw" );
          long i = r.length();
          if (i > 1)
          {
            r.seek( i-1 );
            int ch = r.readUnsignedByte();
            if (ch == 0x1A)
            {
              r.seek( i-1 );
            }
            else
            {
              r.seek( i );
            }
          }
          f.Handle = r;
        }
        else
        {
          f.Handle = null;
        }
      }
      catch (java.lang.SecurityException e)
      {
        f.Handle = null;
      }
      catch (java.io.IOException e)
      {
        f.Handle = null;
      }
      if (f.Handle == null)
      {
        InOutRes = 1;
      }
    }
  );
END;



PROCEDURE Assign( VAR f; id : STRING );
BEGIN
  INLINE
  (
    if (f instanceof SYSTEM_TEXT)
    {
      if (id.length() == 0)
      {
        id = null;
      }
      ((SYSTEM_TEXT)f).FileName = id;
      ((SYSTEM_TEXT)f).Handle = null;
    }
    else if (f instanceof SYSTEM_FILE)
    {
      ((SYSTEM_FILE)f).FileName = id;
      ((SYSTEM_FILE)f).Handle = null;
    }
    else
    {
      InOutRes = 1;
    }
  );
END;


PROCEDURE AssignFile( VAR f; id : STRING );
BEGIN
  Assign( f, id );
END;



PROCEDURE Beep( Freq : LONGWORD; Duration : LONGWORD );
BEGIN
  RAISE( vENotImplemented );
END;



PROCEDURE BlockRead
(
  VAR F      : FILE;
  VAR Buf;
  Count      : LONGWORD;
  VAR Result : LONGWORD
);
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    java.lang.Class c = Buf.getClass();
    if (c.isArray())
    {
      c = c.getComponentType();
      int i = 0;
      try
      {
        if (c == java.lang.Character.TYPE)
        {
          char[] val = (char[])Buf;
          int j = val.length;
          for (i=0; ((i<Count) && (i<j)); i++)
          {
            val[i] = ((java.io.RandomAccessFile)F.Handle).readChar();
          }
        }
        else if (c == java.lang.Boolean.TYPE)
        {
          boolean[] val = (boolean[])Buf;
          int j = val.length;
          for (i=0; ((i<Count) && (i<j)); i++)
          {
            val[i] = ((java.io.RandomAccessFile)F.Handle).readBoolean();
          }
        }
        else if (c == java.lang.Byte.TYPE)
        {
          byte[] val = (byte[])Buf;
          int j = val.length;
          for (i=0; ((i<Count) && (i<j)); i++)
          {
            val[i] = ((java.io.RandomAccessFile)F.Handle).readByte();
          }
        }
        else if (c == java.lang.Short.TYPE)
        {
          short[] val = (short[])Buf;
          int j = val.length;
          for (i=0; ((i<Count) && (i<j)); i++)
          {
            val[i] = ((java.io.RandomAccessFile)F.Handle).readShort();
          }
        }
        else if (c == java.lang.Integer.TYPE)
        {
          int[] val = (int[])Buf;
          int j = val.length;
          for (i=0; ((i<Count) && (i<j)); i++)
          {
            val[i] = ((java.io.RandomAccessFile)F.Handle).readInt();
          }
        }
        else if (c == java.lang.Long.TYPE)
        {
          long[] val = (long[])Buf;
          int j = val.length;
          for (i=0; ((i<Count) && (i<j)); i++)
          {
            val[i] = ((java.io.RandomAccessFile)F.Handle).readLong();
          }
        }
        else if (c == java.lang.Float.TYPE)
        {
          float[] val = (float[])Buf;
          int j = val.length;
          for (i=0; ((i<Count) && (i<j)); i++)
          {
            val[i] = ((java.io.RandomAccessFile)F.Handle).readFloat();
          }
        }
        else if (c == java.lang.Double.TYPE)
        {
          double[] val = (double[])Buf;
          int j = val.length;
          for (i=0; ((i<Count) && (i<j)); i++)
          {
            val[i] = ((java.io.RandomAccessFile)F.Handle).readDouble();
          }
        }
        else if (!c.isPrimitive())
        {
          java.lang.Object[] val = (java.lang.Object[])Buf;
          int j = val.length;
          int[] res = new int[1];
          for (i=0; ((i<Count) && (i<j) && (InOutRes==0)); i++)
          {
            BlockRead( F, val[i], 1, res );
          }
        }
        else
        {
          c = null;
        }
      }
      catch (java.io.EOFException e)
      {
        c = null;
      }
      catch (java.io.IOException e)
      {
        c = null;
      }
      catch (java.lang.RuntimeException e)
      {
        c = null;
      }
      Result[0] = i;
    }
    else if (Count != 1)
    {
      c = null;
      Result[0] = 0;
    }
    else
    {
      Result[0] = 0;
      java.io.FileDescriptor fd;
      java.io.FileInputStream in;
      java.io.ObjectInputStream oi;
      java.lang.Object o;
      try
      {
        fd = ((java.io.RandomAccessFile)F.Handle).getFD();
        in = new java.io.FileInputStream( fd );
        oi = new java.io.ObjectInputStream( in );
        o = oi.readObject();
      }
      catch (java.lang.Exception e)
      {
        c = null;
        o = null;
      }
      if ((o != null) && (Buf.getClass() == o.getClass()))
      {
        java.lang.reflect.Method m;
        java.lang.Class[] paramType = new java.lang.Class[1];
        paramType[0] = Buf.getClass();
        try
        {
          m = Buf.getClass().getDeclaredMethod( "_copy", paramType );
        }
        catch (java.lang.Exception e)
        {
          c = null;
          m = null;
        }
        if (m != null)
        {
          try
          {
            java.lang.Object[] args = new java.lang.Object[1];
            args[0] = o;
            m.invoke( Buf, args );
            Result[0] = 1;
          }
          catch (java.lang.Exception e)
          {
            c = null;
            Result[0] = 0;
          }
        }
        else
        {
          c = null;
        }
      }
      else
      {
        c = null;
      }
    }
    if (c == null)
    {
      InOutRes = 1;
    }
  );
END;


PROCEDURE BlockWrite
(
  VAR F      : FILE;
  VAR Buf;
  Count      : LONGWORD;
  VAR Result : LONGWORD
);
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    java.lang.Class c = Buf.getClass();
    if (c.isArray())
    {
      c = c.getComponentType();
      int i = 0;
      try
      {
        if (c == java.lang.Character.TYPE)
        {
          char[] val = (char[])Buf;
          int j = val.length;
          for (i=0; ((i<Count) && (i<j)); i++)
          {
            ((java.io.RandomAccessFile)F.Handle).writeChar(val[i]);
          }
        }
        else if (c == java.lang.Boolean.TYPE)
        {
          boolean[] val = (boolean[])Buf;
          int j = val.length;
          for (i=0; ((i<Count) && (i<j)); i++)
          {
            ((java.io.RandomAccessFile)F.Handle).writeBoolean(val[i]);
          }
        }
        else if (c == java.lang.Byte.TYPE)
        {
          byte[] val = (byte[])Buf;
          int j = val.length;
          for (i=0; ((i<Count) && (i<j)); i++)
          {
            ((java.io.RandomAccessFile)F.Handle).writeByte(val[i]);
          }
        }
        else if (c == java.lang.Short.TYPE)
        {
          short[] val = (short[])Buf;
          int j = val.length;
          for (i=0; ((i<Count) && (i<j)); i++)
          {
            ((java.io.RandomAccessFile)F.Handle).writeShort(val[i]);
          }
        }
        else if (c == java.lang.Integer.TYPE)
        {
          int[] val = (int[])Buf;
          int j = val.length;
          for (i=0; ((i<Count) && (i<j)); i++)
          {
            ((java.io.RandomAccessFile)F.Handle).writeInt(val[i]);
          }
        }
        else if (c == java.lang.Long.TYPE)
        {
          long[] val = (long[])Buf;
          int j = val.length;
          for (i=0; ((i<Count) && (i<j)); i++)
          {
            ((java.io.RandomAccessFile)F.Handle).writeLong(val[i]);
          }
        }
        else if (c == java.lang.Float.TYPE)
        {
          float[] val = (float[])Buf;
          int j = val.length;
          for (i=0; ((i<Count) && (i<j)); i++)
          {
            ((java.io.RandomAccessFile)F.Handle).writeFloat(val[i]);
          }
        }
        else if (c == java.lang.Double.TYPE)
        {
          double[] val = (double[])Buf;
          int j = val.length;
          for (i=0; ((i<Count) && (i<j)); i++)
          {
            ((java.io.RandomAccessFile)F.Handle).writeDouble(val[i]);
          }
        }
        else if (!c.isPrimitive())
        {
          java.lang.Object[] val = (java.lang.Object[])Buf;
          int j = val.length;
          int[] res = new int[1];
          for (i=0; ((i<Count) && (i<j) && (InOutRes==0)); i++)
          {
            BlockWrite( F, val[i], 1, res );
            if (res[0] != 1)
            {
              break;
            }
          }
        }
        else
        {
          c = null;
        }
      }
      catch (java.io.EOFException e)
      {
        c = null;
      }
      catch (java.io.IOException e)
      {
        c = null;
      }
      catch (java.lang.RuntimeException e)
      {
        c = null;
      }
      Result[0] = Count;
    }
    else if (Count != 1)
    {
      c = null;
      Result[0] = 0;
    }
    else
    {
      Result[0] = 0;
      try
      {
        java.io.FileDescriptor fd = ((java.io.RandomAccessFile)F.Handle).getFD();
        java.io.FileOutputStream out = new java.io.FileOutputStream( fd );
        java.io.ObjectOutputStream oo = new java.io.ObjectOutputStream( out );
        oo.writeObject( Buf );
        Result[0] = 1;
      }
      catch (java.lang.Exception e)
      {
        c = null;
        Result[0] = 0;
      }
    }
    if (c == null)
    {
      InOutRes = 1;
    }
  );
END;




PROCEDURE ByteMove( VAR Source; VAR Dest; Size : LONGWORD );
BEGIN
  RAISE( vENotImplemented );
END;



PROCEDURE ChDir( S : STRING );
BEGIN
  RAISE( vENotImplemented );
END;



PROCEDURE Close( VAR F );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    if (F instanceof SYSTEM_FILE)
    {
      SYSTEM_FILE f = (SYSTEM_FILE)F;
      if (f.Handle != null)
      {
//      try
//      {
//        ((java.io.RandomAccessFile)f.Handle).close();
//        f.Handle = null;
//      }
//      catch (java.lang.Exception e)
//      {
//        InOutRes = 1;
//      }
      }
    }
//  else if (F instanceof SYSTEM_TEXT)
//  {
//    SYSTEM_TEXT f = (SYSTEM_TEXT)F;
//    if (f.Handle != null)
//    {
//      try
//      {
//        ((java.io.RandomAccessFile)f.Handle).close();
//        f.Handle = null;
//      }
//      catch (java.lang.Exception e)
//      {
//        InOutRes = 1;
//      }
//    }
//  }
  );
END;



PROCEDURE CloseFile( VAR F );
BEGIN
  Close( F );
END;




PROCEDURE ClrScr;
BEGIN
  RAISE( vENotImplemented );
END;



FUNCTION Concat( s1, s2 : STRING ) : STRING;
BEGIN
  INLINE
  (
    Result = new java.lang.StringBuffer( s1.toString() + s2.toString() );
//  Result = new java.lang.StringBuffer
//  ( (new java.lang.StringBuffer( s1.toString() )).
//     append( s2.toString() ).toString()
//  );
  );
END;



FUNCTION  Copy( S : STRING; i : INTEGER; Count : INTEGER ) : STRING;
BEGIN
  INLINE
  (
    i -= 1;
    java.lang.String str = new java.lang.String( S );
    int j = i + Count;
    if (i >= str.length())
    {
      Result = new java.lang.StringBuffer( "" );
    }
    else if (j >= str.length())
    {
      Result = new java.lang.StringBuffer( str.substring( i ) );
    }
    else
    {
      Result = new java.lang.StringBuffer( str.substring( i, j ) );
    }
  );
END;



PROCEDURE Delete( VAR S : STRING; i : INTEGER; Count : INTEGER );
BEGIN
  INLINE
  (
    i -= 1;
    java.lang.String str = new java.lang.String( S );
    int j = i + Count;
    java.lang.StringBuffer Result;
    if (i >= str.length())
    {
      Result = S;
    }
    else if (j >= str.length())
    {
      Result = new java.lang.StringBuffer( str.substring( 0, i ) );
    }
    else
    {
      str = str.substring( 0, i ) + str.substring( j, str.length() );
      Result = new java.lang.StringBuffer( str );
    }
    S.setLength(0);
    S.append( Result );
  );
END;



FUNCTION StdinEof : BOOLEAN;
CONST
  EOT = CHR(4);
VAR
  ch : CHAR;
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    Result := FALSE;
    EXIT;
  END;
  INLINE
  (
    try
    {
      if (lookAhead == -1)
      {
        lookAhead = stdin.read();
        if (lookAhead == -1)
        {
          lookAhead = (int)EOT;
        }
      }
      ch = (char)lookAhead;
    }
    catch (java.io.IOException ex)
    {
      ch = (char)0;
    }
  );
  IF (ch = CHR(0)) THEN BEGIN
    Result := FALSE;
    InOutRes := 1;
  END
  ELSE BEGIN
    Result := (ch = EOT);
  END
END;


FUNCTION TextEof( VAR f : TEXT ) : BOOLEAN;
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    if ((f.Handle == null) && (f.FileName == null))
    {
      Result = StdinEof();
    }
    else if (f.Handle == null)
    {
      Result = false;
      InOutRes = 1;
    }
    else
    {
      boolean ok = true;
      try
      {
        long i = ((java.io.RandomAccessFile)f.Handle).getFilePointer();
        long j = ((java.io.RandomAccessFile)f.Handle).length();
        Result = (i >= j);
      }
      catch (java.io.IOException e)
      {
        ok = false;
        Result = false;
      }
      if (!ok)
      {
        Result = false;
        InOutRes = 1;
      }
    }
  );
END;



FUNCTION FileEof( VAR f : FILE ) : BOOLEAN;
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    if (f.Handle == null)
    {
      Result = false;
      InOutRes = 1;
    }
    else
    {
      boolean ok = true;
      try
      {
        long i = ((java.io.RandomAccessFile)f.Handle).getFilePointer();
        long j = ((java.io.RandomAccessFile)f.Handle).length();
        Result = (i >= j);
      }
      catch (java.io.IOException e)
      {
        ok = false;
        Result = false;
      }
      if (!ok)
      {
        Result = false;
        InOutRes = 1;
      }
    }
  );
END;



PROCEDURE Erase( VAR f );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok = false;
    if (f instanceof SYSTEM_FILE)
    {
      SYSTEM_FILE F = (SYSTEM_FILE)f;
      if ((F.FileName != null) && (F.FileName.length() > 0))
      {
        java.io.File file = new java.io.File( F.FileName.toString() );
        try
        {
          ok = file.delete();
        }
        catch (java.lang.SecurityException e)
        {
          ok = false;
        }
      }
    }
    else if (f instanceof SYSTEM_TEXT)
    {
      SYSTEM_TEXT F = (SYSTEM_TEXT)f;
      if ((F.FileName != null) && (F.FileName.length() > 0))
      {
        java.io.File file = new java.io.File( F.FileName.toString() );
        try
        {
          ok = file.delete();
        }
        catch (java.lang.SecurityException e)
        {
          ok = false;
        }
      }
    }
    else
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;



FUNCTION FilePos( VAR f : FILE ) : LONGINT;
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok;
    if (f.Handle == null)
    {
      ok = false;
    }
    else
    {
      try
      {
        Result = (int)((java.io.RandomAccessFile)f.Handle).getFilePointer();
        ok = true;
      }
      catch (java.io.IOException e)
      {
        ok = false;
      }
    }
    if (!ok)
    {
      Result = -1;
      InOutRes = 1;
    }
  );
END;



FUNCTION FileSize( VAR f : FILE ) : LONGINT;
BEGIN
  (* Note:
     Canterbury Pascal always returns the file length in bytes.
     Borland Object Pascal returns the file size in bytes,
     unless it is a file of records, in which case it returns
     the number of records.
  *)
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok;
    if (f.Handle == null)
    {
      ok = false;
    }
    else
    {
      try
      {
        Result = (int)((java.io.RandomAccessFile)f.Handle).length();
        ok = true;
      }
      catch (java.io.IOException e)
      {
        ok = false;
      }
    }
    if (!ok)
    {
      Result = -1;
      InOutRes = 1;
    }
  );
END;



PROCEDURE FillChar( VAR Dest; Size : INTEGER; Value : INTEGER );
BEGIN
  INLINE
  (
    boolean ok;
    java.lang.Class c = Dest.getClass();
    if (c.isArray())
    {
      java.lang.Class t = c.getComponentType();
      if (t == java.lang.Byte.TYPE)
      {
        byte[] b = (byte[])Dest;
        int j = b.length;
        int i;
        if (Size < j)
        {
          j = Size;
        }
        for (i=0; i<j; i++)
        {
          b[i] = (byte)Value;
        }
      }
      else if (t == java.lang.Character.TYPE)
      {
        char[] b = (char[])Dest;
        int j = b.length;
        int i;
        if (Size < j)
        {
          j = Size;
        }
        for (i=0; i<j; i++)
        {
          b[i] = (char)Value;
        }
      }
      else
      {
        throw new SYSTEM_EInvalidCast();
      }
    }
  );
END;



PROCEDURE GetDir( Drive : BYTE; VAR S : STRING );
BEGIN
  (* Get current working directory.

     Note: Borland Object Pascal accepts non-zero values
           for 'Drive'.
           Canterbury Pascal only accepts zero for 'Drive',
           meaning that it can only find the current working
           directory for the currently active drive.
  *)
  INLINE
  (
    java.io.File f = new java.io.File( "." );
    java.lang.String s;

    S.setLength(0);
    try
    {
      S.append( (Drive != (byte)0) ? "." : f.getCanonicalPath() );
    }
    catch (java.io.IOException e)
    {
      S.setLength(0);
      S.append( "." );
    }
  );
END;



PROCEDURE GotoXY( X : BYTE; Y : BYTE );
BEGIN
  RAISE( vENotImplemented );
END;



PROCEDURE HALT( code : BYTE );
BEGIN
  INLINE
  (
    java.lang.System.runFinalizersOnExit( true );
    java.lang.System.runFinalization();
    java.lang.System.exit( code );
  );
END;



PROCEDURE CANCEL( e : POINTER );
BEGIN
  INLINE
  (
    ((java.lang.Exception)e).printStackTrace();
    if (args == null)
    {
      throw new java.lang.RuntimeException();
    }
    else
    {
      HALT( (byte)1 );
    }
  );
END;




PROCEDURE Insert( Source : STRING; VAR Dest : STRING; i : INTEGER );
BEGIN
  INLINE
  (
    i -= 1;
    java.lang.String s = new java.lang.String();
    try
    {
      s = Dest.toString().substring( 0, i );
      s += Source.toString();
      s += Dest.toString().substring( i );
    }
    catch (java.lang.StringIndexOutOfBoundsException e)
    {
      s = "";
    }
    Dest.setLength(0);
    Dest.append(s);
  );
END;



FUNCTION KeyPressed : BOOLEAN;
BEGIN
  INLINE
  (
    try
    {
      Result = (java.lang.System.in.available() > 0);
    }
    catch (java.io.IOException e)
    {
      Result = false;
    }
  );
END;



FUNCTION Length( S : STRING ) : INTEGER;
BEGIN
  INLINE
  (
    Result = S.length();
  );
END;



PROCEDURE MainDispatchLoop;
BEGIN
  RAISE( vENotImplemented );
END;



FUNCTION MaxAvail : LONGWORD;
BEGIN
  INLINE
  (
    long i = java.lang.Runtime.getRuntime().totalMemory();
    Result = ((i & 0xFFFFFFFF00000000L) != 0L) ? 0xFFFFFFFF : (int)i;
  );
END;



FUNCTION MemAvail : LONGWORD;
BEGIN
  INLINE
  (
    long i = java.lang.Runtime.getRuntime().freeMemory();
    Result = ((i & 0xFFFFFFFF00000000L) != 0L) ? 0xFFFFFFFF : (int)i;
  );
END;



PROCEDURE MkDir( S : STRING );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok;
    try
    {
      java.io.File f = new java.io.File( S.toString() );
      ok = f.mkdir();
    }
    catch (java.lang.SecurityException e)
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;



PROCEDURE Move( VAR Source; VAR Dest; Count : INTEGER );
BEGIN
  INLINE
  (
    int Size = Count;
    boolean ok;
    java.lang.Class c1 = Source.getClass();
    java.lang.Class c2 = Dest.getClass();
    if ((c1.isArray()) && (c2.isArray()))
    {
      java.lang.Class t1 = c1.getComponentType();
      java.lang.Class t2 = c2.getComponentType();
      if ( (t1 == java.lang.Byte.TYPE) &&
           (t2 == java.lang.Byte.TYPE)
         )
      {
        byte[] b1 = (byte[])Source;
        byte[] b2 = (byte[])Dest;
        int j1 = b1.length;
        int j2 = b2.length;
        int i;
        if (Size > j1)
        {
          Size = j1;
        }
        if (Size > j2)
        {
          Size = j2;
        }
        java.lang.System.arraycopy( b1, 0, b2, 0, Size );
      }
      else if ( (t1 == java.lang.Character.TYPE) &&
                (t2 == java.lang.Character.TYPE)
              )
      {
        char[] b1 = (char[])Source;
        char[] b2 = (char[])Dest;
        int j1 = b1.length;
        int j2 = b2.length;
        int i;
        if (Size > j1)
        {
          Size = j1;
        }
        if (Size > j2)
        {
          Size = j2;
        }
        java.lang.System.arraycopy( b1, 0, b2, 0, Size );
      }
      else
      {
        throw new SYSTEM_EInvalidCast();
      }
    }
  );
END;



FUNCTION ParamCount : INTEGER;
BEGIN
  INLINE
  (
    if (args != null)
    {
      java.lang.String[] s = (java.lang.String[])args;
      Result = s.length;
    }
    else
    {
      Result = 0;
    }
  );
END;



FUNCTION ParamStr( N : INTEGER ) : STRING;
BEGIN
  INLINE
  (
    if ((args != null) && (N > 0))
    {
      java.lang.String[] s = (java.lang.String[])args;
      Result.setLength(0);
      if (N <= s.length)
      {
        Result.append( s[N-1] );
      }
    }
    else
    {
      Result.setLength(0);
    }
  );
END;



FUNCTION Pos( SubStr : STRING; S : STRING ) : INTEGER;
BEGIN
  INLINE
  (
    Result = (byte)S.toString().indexOf( SubStr.toString() ) + 1;
  );
END;




FUNCTION Random : REAL;
BEGIN
  (*@IF IsLongReal *)
  INLINE
  (
    Result = java.lang.Math.random();
  );
  (*@ELSE *)
  INLINE
  (
    Result = (float)java.lang.Math.random();
  );
  (*@ENDIF *)
END;




PROCEDURE Randomize;
BEGIN
  { obsolete }
END;



PROCEDURE Rename( VAR F; NewName : STRING );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok;
    long i;
    java.lang.String OldName;
    if (F instanceof SYSTEM_FILE)
    {
      OldName = ((SYSTEM_FILE)F).FileName.toString();
    }
    else if (F instanceof SYSTEM_TEXT)
    {
      if (((SYSTEM_TEXT)F).FileName != null)
      {
        OldName = ((SYSTEM_TEXT)F).FileName.toString();
      }
      else
      {
        OldName = "";
      }
    }
    else
    {
      OldName = null;
    }
    if (OldName != null)
    {
      if (NewName.toString().equals( "" ))
      {
        ok = false;
      }
      else
      {
        try
        {
          java.io.File oldFile = new java.io.File( OldName );
          java.io.File newFile = new java.io.File( NewName.toString() );
          if (oldFile.renameTo(newFile))
          {
            ok = true;
          }
          else
          {
            ok = false;
          }
        }
        catch (java.lang.SecurityException e)
        {
          ok = false;
        }
      }
      if (ok)
      {
        if (F instanceof SYSTEM_FILE)
        {
          SYSTEM_FILE f = (SYSTEM_FILE)F;
          f.FileName = NewName;
          f.Handle = null;
        }
        else if (F instanceof SYSTEM_TEXT)
        {
          SYSTEM_TEXT t = (SYSTEM_TEXT)F;
          t.FileName = NewName;
          t.Handle = null;
        }
        else
        {
          ok = false;
        }
      }
    }
    else
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;



PROCEDURE Reset( VAR F );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok = true;
    java.lang.String name;
    java.lang.StringBuffer id;
    if (F instanceof SYSTEM_FILE)
    {
      id = ((SYSTEM_FILE)F).FileName;
    }
    else if (F instanceof SYSTEM_TEXT)
    {
      id = ((SYSTEM_TEXT)F).FileName;
    }
    else
    {
      id = null;
    }
    name = (id != null) ? id.toString() : null;
    if ((name != null) && (name.length() > 0))
    {
      java.io.File f = new java.io.File( new java.lang.String( name ) );
      try
      {
        if ((!f.exists()) || (!f.isFile()))
        {
          ok = false;
        }
      }
      catch ( java.lang.SecurityException e )
      {
        ok = false;
      };
      if (ok)
      {
        java.io.RandomAccessFile r;
        try
        {
          if (F instanceof SYSTEM_FILE)
          {
            r = new java.io.RandomAccessFile( name, "rw" );
          }
          else
          {
            r = new java.io.RandomAccessFile( name, "r" );
          }
        }
        catch (java.io.IOException e)
        {
          r = null;
          ok = false;
        }
        catch (java.lang.SecurityException e)
        {
          r = null;
          ok = false;
        }
        if (F instanceof SYSTEM_FILE)
        {
          ((SYSTEM_FILE)F).Handle = r;
        }
        else if (F instanceof SYSTEM_TEXT)
        {
          ((SYSTEM_TEXT)F).Handle = r;
        }
        else
        {
          ok = false;
        }
      }
    }
    else if (F instanceof SYSTEM_TEXT)
    {
      ok = true;
    }
    else
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;



PROCEDURE Rewrite( VAR F );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok = true;
    java.lang.String name;
    java.lang.StringBuffer id;
    java.io.RandomAccessFile r;
    java.io.File f;
    if (F instanceof SYSTEM_FILE)
    {
      id = ((SYSTEM_FILE)F).FileName;
    }
    else if (F instanceof SYSTEM_TEXT)
    {
      id = ((SYSTEM_TEXT)F).FileName;
    }
    else
    {
      id = null;
    }
    name = (id != null) ? id.toString() : null;
    if ((name != null) && (name.length() > 0))
    {
      try
      {
        if ((F instanceof SYSTEM_FILE) && (((SYSTEM_FILE)F).Handle != null))
        {
          ((java.io.RandomAccessFile)((SYSTEM_FILE)F).Handle).close();
          ((SYSTEM_FILE)F).Handle = null;
        }
        else if ((F instanceof SYSTEM_TEXT) && (((SYSTEM_TEXT)F).Handle != null))
        {
          ((java.io.RandomAccessFile)((SYSTEM_TEXT)F).Handle).close();
          ((SYSTEM_TEXT)F).Handle = null;
        }
        f = new java.io.File( name );
        if (f.exists())
        {
          if (f.isFile())
          {
            f.delete();
          }
        }
        r = new java.io.RandomAccessFile( name, "rw" );
      }
      catch (java.io.IOException e)
      {
        r = null;
        ok = false;
      }
      catch (java.lang.SecurityException e)
      {
        r = null;
        ok = false;
      }
      if (F instanceof SYSTEM_FILE)
      {
        ((SYSTEM_FILE)F).Handle = r;
      }
      else if (F instanceof SYSTEM_TEXT)
      {
        ((SYSTEM_TEXT)F).Handle = r;
      }
      else
      {
        ok = false;
      }
    }
    else if (F instanceof SYSTEM_TEXT)
    {
      ok = true;
    }
    else
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;



PROCEDURE RmDir( S : STRING );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok;
    try
    {
      ok = false;
      java.io.File f = new java.io.File( S.toString() );
      if (f.isDirectory())
      {
        if (f.delete())
        {
          ok = true;
        }
      }
    }
    catch (java.lang.SecurityException e)
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;



PROCEDURE RunError( retcode : BYTE );
BEGIN
  INLINE
  (
    if (args == null)
    {
      throw new java.lang.RuntimeException();
    }
    else
    {
      HALT( (byte)1 );
    }
  );
END;



PROCEDURE Seek( VAR F : FILE; N : LONGINT );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok = false;
    if (F.Handle != null)
    {
      try
      {
        ((java.io.RandomAccessFile)F.Handle).seek( N & 0xFFFFFFFFL );
        ok = true;
      }
      catch (java.io.IOException e)
      {
        ok = false;
      }
    }
    else
    {
      ok = false;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;




PROCEDURE IntToStr( i : INTEGER; VAR s : STRING; prec : INTEGER );
BEGIN
  INLINE
  (
    java.lang.String str = java.lang.Integer.toString( i );
    int j = str.length();
    if (prec == 0)
    {
      prec = j;
    }
    while (j < prec)
    {
      str = " " + str;
      j += 1;
    }
    while ((j > prec) && (str.charAt(0) == ' '))
    {
      str = str.substring(1);
      j -= 1;
    }
    if (j > prec)
    {
      throw new SYSTEM_EConvertError();
    }
    s.setLength(0);
    s.append(str);
  );
END;



PROCEDURE StringsMemCpy
(
  Dest      : POINTER;
  DestPos   : LONGINT;
  Source    : POINTER;
  SourcePos : LONGINT;
  Size      : LONGINT
);
BEGIN
  INLINE
  (
    java.lang.System.arraycopy( Source, SourcePos, Dest, DestPos, Size );
  );
END;



PROCEDURE StringsInsert
( SubStr           : ARRAY OF CHAR;
  VAR Str          : ARRAY OF CHAR;
  Index            : CARDINAL
);
FORWARD;



PROCEDURE StringsAssign
( VAR Source       : ARRAY OF CHAR;
  VAR Destination  : ARRAY OF CHAR
);
VAR
  i,j              : CARDINAL;
BEGIN
  i := StringsLength( Source );
  j := HIGH( Destination )+1;
  IF i > j THEN BEGIN
    i := j;
  END;
  StringsMemCpy( ADDR( Destination ), 0, ADDR( Source ), 0, i );
  IF i <= HIGH( Destination ) THEN BEGIN
    Destination[ i ] := CHR(0);
  END;
END;



PROCEDURE StringsConcat
( Str1             : ARRAY OF CHAR;
  Str2             : ARRAY OF CHAR;
  VAR Result       : ARRAY OF CHAR
);
BEGIN
  StringsAssign(Str1,Result);
  StringsInsert(Str2,Result,StringsLength(Str1));
END;



PROCEDURE StringsInsert
( SubStr           : ARRAY OF CHAR;
  VAR Str          : ARRAY OF CHAR;
  Index            : CARDINAL
);
VAR
  i,j,k            : CARDINAL;
BEGIN
  i := StringsLength( SubStr );
  j := StringsLength( Str );
  IF Index > j THEN BEGIN
    StringsConcat( Str, SubStr, Str );
  END
  ELSE BEGIN
    IF i+j <= HIGH(Str) THEN BEGIN
      Str[i+j] := CHR(0);
    END (* if *);
    IF i > HIGH( Str ) + 1 - Index THEN BEGIN
      i := HIGH( Str ) + 1 - Index;
      StringsMemCpy( ADDR( Str ), Index, ADDR( SubStr ), 0, i );
    END
    ELSE BEGIN (* Index + Length( SubStr ) <= HIGH( Str ) + 1 *)
      k := j - Index;
      IF (i+j) > (HIGH(Str)+1) THEN BEGIN
        k := HIGH( Str ) + 1 - Index - i;
      END (* if *);
      StringsMemCpy( ADDR( Str ), Index+i, ADDR( Str ), Index, k );
      StringsMemCpy( ADDR( Str ), Index, ADDR( SubStr ), 0, i );
    END;
  END;
END;


PROCEDURE StringsAppend
( SubStr           : ARRAY OF CHAR;
  VAR Str          : ARRAY OF CHAR
);
BEGIN
  StringsInsert( SubStr, Str, StringsLength( Str ) );
END;



PROCEDURE StringsCopy( source : ARRAY OF CHAR; VAR dest : ARRAY OF CHAR );
BEGIN
  dest[0] := CHR(0);
  StringsAppend( source, dest );
END;



PROCEDURE StringsFill
( Pos              : LONGWORD;
  Size             : LONGWORD;
  Filler           : CHAR;
  VAR Str          : ARRAY OF CHAR
);
VAR
  i                : LONGWORD;
  j                : LONGWORD;
  highStr          : LONGWORD;
BEGIN
  highStr := HIGH(Str);
  i := StringsLength( Str );
  IF Pos > i THEN BEGIN
    EXIT;
  END;
  IF Pos = i THEN BEGIN
    WHILE (i < highStr) AND (Size > 0) DO BEGIN
      Str[ i ] := Filler;
      INC( i );
      DEC( Size );
    END;
    Str[ i ] := CHR(0);
    EXIT;
  END;
  IF Pos + Size >= highStr THEN BEGIN
    Size := highStr - Pos;
    WHILE Size > 0 DO BEGIN
      Str[ Pos ] := Filler;
      INC( Pos );
      DEC( Size );
    END;
    Str[ Pos ] := CHR(0);
    EXIT;
  END;
  j := i - Pos;
  IF Pos + Size + j >= highStr THEN BEGIN
    j := highStr - Pos - Size;
  END;
  i := Pos + Size + j;
  Str[ i ] := CHR(0);
  WHILE j > 0 DO BEGIN
    DEC( j );
    DEC( i );
    Str[ i ] := Str[ i-Size ];
  END;
  i := Pos;
  WHILE i < Pos+Size DO BEGIN
    Str[ i ] := Filler;
    INC( i );
  END;
END;




PROCEDURE floatcvt
(
  isLong      : BOOLEAN;
  cnvflag     : BOOLEAN;
  r           : DOUBLE;
  digits      : SMALLINT;
  VAR decpt   : SMALLINT;
  VAR sign    : BOOLEAN;
  VAR result  : DIGITSTRING;
  VAR done    : BOOLEAN
);
VAR
  i           : SMALLINT;
  pow         : SMALLINT;
  digstr      : DIGITSTRING;
  p           : SMALLINT;
  n           : LONGINT;
  digprec     : SMALLINT;
  highResult  : LONGINT;
BEGIN
(*
   Convert a long real value to a string of decimal digits.
   Input:
     isLong    = up to (15 if TRUE; 6 if FALSE) digits are significand
     cnvflag = how to interpret wanted # of digits
     r       = long real value to be converted
     digits  = if cnvflag = FALSE:
                 # of digits in resulting string past the decimal point
               if cnvflag = TRUE:
                 # of digits in resulting string
   Returns:
     decpt  = position of decimal point from left of first digit
     sign   = TRUE if value was negative
     result = string with digits
   BUGS:
     This routine will hang if it is passed a NAN or INFINITY.
*)
  highResult := HIGH(result);
  IF r < 0.0 THEN BEGIN
    sign := TRUE;
    r := -r;
  END
  ELSE BEGIN
    sign := FALSE;
  END;
  IF digits < 0 THEN BEGIN
    digits := 0;
  END
  ELSE IF digits >= DIGMAX THEN BEGIN
    digits := DIGMAX;
  END;
  IF r = 0.0 THEN BEGIN
    result[ 0 ] := CHR(0);
    StringsFill( 0, highResult, '0', result );
    decpt := 0;
    done := TRUE;
    EXIT;
  END;

  (* Adjust things so that 1 <= r < 10 *)
  decpt := 1;
  pow := 256;
  i := 0;
  WHILE r < 1. DO BEGIN
    WHILE r < negtab[ i+1 ] DO BEGIN
      r := r * postab[ i ];
      decpt := decpt - pow;
    END;
    pow := pow DIV 2;
    i := i + 1;
  END;
  pow := 256;
  i := 0;
  WHILE r >= 10. DO BEGIN
    WHILE r >= postab[ i ] DO BEGIN
      r := r * negtab[ i ];
      decpt := decpt + pow;
    END;
    pow := pow DIV 2;
    i := i + 1;
  END;

  IF cnvflag THEN BEGIN
    digits := digits + decpt;
    IF digits < 0 THEN BEGIN
      digits := 0;
    END
    ELSE IF digits >= DIGMAX THEN BEGIN
      digits := DIGMAX;
    END;
  END;

  (* Pick off digits 1 by 1 and stuff into digstr[] *)
  (* Do 1 extra digit for rounding purposes         *)
  IF isLong THEN BEGIN
    digprec := DBLDIG+1;
  END
  ELSE BEGIN
    digprec := SGLDIG+1;
  END;
  FOR p := 0 TO digits DO BEGIN
    digstr[p] := CHR(0);
    IF p >= digprec THEN BEGIN
      (* only 'digprec' digits are significand *)
      digstr[ p ] := '0';
    END
    ELSE BEGIN
      (* store digit before getting next one *)
      n := TRUNC( r );
      digstr[ p ] := CHR( n + ORD('0') );
      r := (r - LONGFLOAT(n)) * 10.0;
    END;
  END;

  IF digits < digprec THEN BEGIN
    p := digits;
  END
  ELSE BEGIN
    p := digprec-1;
  (*p := digprec;*)
  END;
  IF digstr[ p ] >= '5' THEN BEGIN
    (* we need to round *)
    WHILE TRUE DO BEGIN
      IF p = 0 THEN BEGIN (* if at start *)
        IF cnvflag THEN BEGIN
          digits := digits + 1;
        END;
        decpt := decpt + 1; (* shift dec pnt *)
        digstr[ 0 ] := '1'; (* "100000..."   *)
        BREAK;
      END;
      digstr[ p ] := '0';
      p := p - 1;
      IF digstr[ p ] <> '9' THEN BEGIN
        digstr[p] := CHR( ORD( digstr[p] )+1 );
        BREAK;
      END;
    END;
  END;

  digstr[ digits ] := CHR(0);  (* terminate string *)
  StringsCopy( digstr, result );
  done := TRUE;
END;



PROCEDURE FloatToFixPt
(
  isLong   : BOOLEAN;
  r        : DOUBLE;
  size     : SMALLINT;
  decimals : SMALLINT;
  VAR s    : ARRAY OF CHAR;
  VAR done : BOOLEAN
);
VAR
  i       : SMALLINT;
  p       : SMALLINT;
  decpt   : SMALLINT;
  sign    : BOOLEAN;
  buf     : DIGITSTRING;
  j       : SMALLINT;
  ch      : CHAR;
BEGIN
  (*
     Convert a long real value into a fixed floating point format:
     ["-"] digit {digit} ["." digit {digit}]
  *)

  IF IsLongINFINITY(r) THEN BEGIN
    StringsCopy( 'infinity', s );
    done := FALSE;
    EXIT;
  END
  ELSE IF IsLongNAN(r) THEN BEGIN
    StringsCopy( 'NaN', s );
    done := FALSE;
    EXIT;
  END;

  (* make sure 'size' and 'decimals' are in proper ranges *)
  IF r < 0. THEN BEGIN
    i := 1;
  END
  ELSE BEGIN
    i := 0;
  END;
  IF size > HIGH( s ) THEN BEGIN
    (* size does not fit into string *)
    s[0] := CHR(0);
    done := FALSE;
    EXIT;
  END;
  IF decimals > 0 THEN BEGIN
    j := 1;
  END
  ELSE BEGIN
    j := 0;
  END;
  IF size < i + 1 + j THEN BEGIN
    (* at least ["-"] digit ["."] needed for result string *)
    s[0] := CHR(0);
    done := FALSE;
    EXIT;
  END;
  IF decimals > DIGMAX THEN BEGIN
    decimals := DIGMAX;
  END
  ELSE IF decimals < 0 THEN BEGIN
    s[0] := CHR(0);
    done := FALSE;
    EXIT;
  END;
  IF size - i - 1 - j < decimals THEN BEGIN
    (* too many decimals wanted after dec. point *)
    s[0] := CHR(0);
    done := FALSE;
    EXIT;
  END;
  s[ size ] := CHR(0);

  (* store significand digits into a buffer *)
  floatcvt( isLong, TRUE, r, decimals, decpt, sign, buf, done );

  p := 0;
  i := 0;
  IF i >= size THEN BEGIN
    done := FALSE;
    EXIT;
  END;
  IF sign THEN BEGIN
    s[i] := '-';
    INC( i );
  END;
  IF i >= size THEN BEGIN
    done := FALSE;
    EXIT;
  END;

  IF decpt <= 0 THEN BEGIN
    (* at least 1 digit before dec point *)
    s[i] := '0';
    INC( i );
  END;
  IF i >= size THEN BEGIN
    done := FALSE;
    EXIT;
  END;

  (* store all digits before dec. point into result string *)
  ch := buf[ p ];
  WHILE (decpt > 0) AND (i < size) AND (ch <> CHR(0)) DO BEGIN
    s[i] := buf[p];
    INC( i );
    INC( p );
    DEC( decpt );
    ch := buf[p];
  END;
  WHILE (decpt > 0) AND (i < size) DO BEGIN
    s[i] := '0';
    INC( i );
    DEC( decpt );
  END;
  IF decpt > 0 THEN BEGIN
    done := FALSE;
    EXIT;
  END;

  (* store the dec. point itself into result string *)
  IF decimals > 0 THEN BEGIN
    s[i] := '.';
    INC( i );
    IF i >= size THEN BEGIN
      done := FALSE;
      EXIT;
    END;
  END;

  (* now do the digits after the dec. point ... *)
  WHILE (decpt < 0) AND (decimals > 0) AND (i < size) DO BEGIN
    s[i] := '0';
    INC( i );
    INC( decpt );
    DEC( decimals );
  END;
  ch := buf[p];
  WHILE (decimals > 0) AND (i < size) AND (ch <> CHR(0)) DO BEGIN
    s[i] := ch;
    DEC( decimals );
    INC( p );
    INC( i );
    ch := buf[p];
  END;
  WHILE (decimals > 0) AND (i < size) DO BEGIN
    s[i] := '0';
    DEC( decimals );
    INC( i );
  END;
  IF decimals > 0 THEN BEGIN
    s[i] := CHR(0);
    done := FALSE;
    EXIT;
  END;

  (* finally insert leading blanks as needed *)
  s[i] := CHR(0);
  IF i < size THEN BEGIN
    j := size - i;
    StringsFill( 0, j, ' ', s );
  END;
END;




PROCEDURE LongRealToFixPt
( r        : DOUBLE;
  size     : SMALLINT;
  digits   : SMALLINT;
  VAR s    : ARRAY OF CHAR;
  VAR done : BOOLEAN
);
BEGIN
  FloatToFixPt( TRUE, r, size, digits, s, done );
END;


PROCEDURE CharArrayToPascalString
(
  VAR source : ARRAY OF CHAR;
  VAR dest   : STRING
);
BEGIN
  INLINE
  (
    int i = 0;
    int j = source.length;
    while ((i < j) && (source[i] != (char)0))
    {
      i++;
    }
    dest.setLength(0);
    dest.append( source, 0, i );
  );
END;


(*@IF FALSE *)


PROCEDURE AdjustExponent( VAR s : STRING; incr : LONGINT; islong : BOOLEAN );
BEGIN
  INLINE
  (
    java.lang.String str = s.toString();
    int i = 0;
    int exponent = 0;
    if (str.charAt(0) == '-')
    {
      i = 1;
    }
    i = str.indexOf( 'E', i );
    if (i > 0)
    {
      if (str.charAt( i+1 ) == '+')
      {
        exponent = java.lang.Integer.parseInt( str.substring(i+2) );
      }
      else
      {
        exponent = java.lang.Integer.parseInt( str.substring(i+1) );
      }
      exponent += incr;
      s.setLength(i+1);
      if (exponent >= 0)
      {
        s.append( '+' );
      }
      else
      {
        s.append( '-' );
        exponent = -exponent;
      }
      if (exponent < 10)
      {
        s.append( '0' );
      }
      if ((exponent < 100) && (islong))
      {
        s.append( '0' );
      }
      s.append( exponent );
    }
  );
END;


PROCEDURE AdjustFraction( VAR s : STRING; decimals : LONGINT );
BEGIN
  INLINE
  (
    java.lang.String str = s.toString();
    int i = str.indexOf( '.' );
    int j = str.length();
    int k = i + 1 + decimals;
    int n;
    if (i >= 0)
    {
      i += 1;
      while ((i < j) && (i < k) && (str.charAt(i) != 'E'))
      {
        i += 1;
      }
      s.setLength(0);
      if (str.charAt(0) != '-')
      {
        s.append( ' ' );
      }
      s.append( str.substring( 0, i ) );
      n = str.indexOf( 'E' );
      while (i < k)
      {
        s.append( '0' );
        i += 1;
      }
      if (n >= 0)
      {
        s.append( str.substring( n ) );
      }
    }
  );
END;


(*@ENDIF *)


PROCEDURE FloatToScientificNotation
(
  isLong   : BOOLEAN;
  r        : DOUBLE;
  digits   : SMALLINT;
  VAR s    : STRING;
  VAR done : BOOLEAN
);
VAR
  p        : LONGINT;
  decpt    : SMALLINT;
  sign     : BOOLEAN;
  buf      : DIGITSTRING;
  ch       : CHAR;
BEGIN
  (* convert a real value to a scientific notation:

     [ |-] digit '.' {digit} E [+|-] digit digit [digit]
  *)

  (* store significand digits into 'buf' *)
  floatcvt( isLong, FALSE, r, digits, decpt, sign, buf, done );

  (* do the significand part: [-] digit . {digit} *)
  s := '';
  p := 0;
  IF sign THEN BEGIN
    s := s + '-';
  END
  ELSE BEGIN
    s := s + ' ';
  END;
  IF digits > 0 THEN BEGIN
    s := s + buf[p];
    INC( p );
    DEC( digits );
  END
  ELSE BEGIN
    s := s + '0';
  END;
  s := s + '.';
  ch := buf[p];
  WHILE (digits > 0) AND (ch > CHR(0)) DO BEGIN
    s := s + ch;
    INC( p );
    ch := buf[p];
    DEC( digits );
  END;
  IF digits > 0 THEN BEGIN
    done := FALSE;
    EXIT;
  END;

  (* now do the exponent part: E (+|-) digit digit digit *)
  IF r <> 0.0 THEN BEGIN (* avoid 0.00e-01 *)
    DEC( decpt );
  END;
  s := s + 'E';
  IF decpt < 0 THEN BEGIN
    s := s + '-';
    decpt := -decpt;
  END
  ELSE BEGIN
    s := s + '+';
  END;
  IF isLong THEN BEGIN
    s := s + CHR( decpt DIV 100 + ORD('0') );
    decpt := decpt MOD 100;
  END;
  s := s + CHR( decpt DIV 10 + ORD('0') );
  decpt := decpt MOD 10;
  s := s + CHR( decpt + ORD('0') );
END;



PROCEDURE RealToStr
(
  r      : DOUBLE;
  VAR s  : STRING;
  prec1  : INTEGER;
  prec2  : INTEGER;
  islong : BOOLEAN
);
VAR
  done   : BOOLEAN;
  str    : ARRAY [0..255] OF CHAR;
  digits : LONGINT;
  n      : LONGINT;
BEGIN
  done := FALSE;
  IF prec1 < 0 THEN BEGIN
    prec1 := 0;
  END;
  IF IsLongINFINITY(r) THEN BEGIN
    IF r < 0.0 THEN BEGIN
      s := '-INF';
    END
    ELSE BEGIN
      s := 'INF';
    END;
    done := TRUE;
  END
  ELSE IF IsLongNAN(r) THEN BEGIN
    s := 'NAN';
    done := TRUE;
  END
  ELSE BEGIN
    IF prec2 >= 0 THEN BEGIN
      (* Fixed point notation wanted *)
      IF prec2 > 11 THEN BEGIN
        (* never use more than 11 digits after decimal point *)
        prec2 := 11;
      END;
      LongRealToFixPt( r, SMALLINT(prec1), SMALLINT(prec2), str, done );
      IF done THEN BEGIN
        CharArrayToPascalString( str, s );
      END
      ELSE BEGIN
        RealToStr( r, s, prec1, -1, islong );
        EXIT;
      END;
    END
    ELSE BEGIN
      (* Floating point notation wanted in the following format:
        [   | - ] <digit> . <decimals> E [ + | - ] <exponent>
      *)
      IF prec1 < 8 THEN BEGIN
        (* Minimum width must be at least 8 characters *)
        prec1 := 8;
      END;
      IF islong THEN BEGIN
        digits := prec1 - 7;
      END
      ELSE BEGIN
        digits := prec1 - 6;
      END;
      IF digits > 11 THEN BEGIN
        digits := 11;
      END;
      FloatToScientificNotation( islong, r, SMALLINT(digits), s, done );
    END;
  END;
  IF NOT done THEN BEGIN
    str[0] := CHR(0);
    StringsFill( 0, prec1, '?', str );
    CharArrayToPascalString( str, s );
  END
  ELSE BEGIN
    (* insert leading blanks as needed to satisfy minimum width *)
    n := prec1 - Length(s);
    WHILE n > 0 DO BEGIN
      s := ' ' + s;
      n := n - 1;
    END;
  END;
END;



FUNCTION ToStr( l : LONGINT ) : STRING;
BEGIN
  INLINE
  (
    Result = new java.lang.StringBuffer( java.lang.Integer.toString(l) );
  );
END;



PROCEDURE Truncate( VAR F : FILE );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;

  INLINE
  (
    boolean ok = false;
    java.io.File f;
    if (F.Handle != null)
    {
      try
      {
        ((java.io.RandomAccessFile)F.Handle).close();
        F.Handle = null;
        f = new java.io.File( new java.lang.String( F.FileName ) );
        f.delete();
        F.Handle = new java.io.RandomAccessFile( f, "rw" );
        ok = true;
      }
      catch ( java.io.IOException e )
      {
        F.Handle = null;
        F.FileName = null;
        ok = false;
      }
      catch ( java.lang.SecurityException e )
      {
        F.Handle = null;
        F.FileName = null;
        ok = false;
      }
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;



FUNCTION UpCase( ch : CHAR ) : CHAR;
BEGIN
  INLINE
  (
    ch = java.lang.Character.toUpperCase(ch);
  );
  UpCase := ch;
END;



FUNCTION IOResult : INTEGER;
BEGIN
  IOResult := InOutRes;
  InOutRes := 0;
END;



PROCEDURE CheckIO;
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    InOutRes := 0;
    RAISE( NEW( EInOutError ) );
  END;
END;



PROCEDURE _Assign( VAR f; id : STRING );
BEGIN
  Assign( f, id );
  CheckIO;
END;



PROCEDURE _AssignFile( VAR f; id : STRING );
BEGIN
  Assign( f, id );
  CheckIO;
END;



PROCEDURE _Append( VAR f : TEXT );
BEGIN
  Append( f );
  CheckIO;
END;


PROCEDURE _Close( VAR F );
BEGIN
  Close( F );
  CheckIO;
END;


PROCEDURE _CloseFile( VAR F );
BEGIN
  Close( F );
  CheckIO;
END;


FUNCTION _TextEof( VAR f : TEXT ) : BOOLEAN;
BEGIN
  Result := TextEof( f );
  CheckIO;
END;



FUNCTION _FileEof( VAR f : FILE ) : BOOLEAN;
BEGIN
  Result := FileEof( f );
  CheckIO;
END;



FUNCTION _StdinEof : BOOLEAN;
BEGIN
  Result := StdinEof;
  CheckIO;
END;



PROCEDURE _Erase( VAR f );
BEGIN
  Erase( f );
  CheckIO;
END;



FUNCTION _FilePos( VAR f : FILE ) : LONGINT;
BEGIN
  Result := FilePos( f );
  CheckIO;
END;



FUNCTION _FileSize( VAR f : FILE ) : LONGINT;
BEGIN
  Result := FileSize( f );
  CheckIO;
END;




PROCEDURE _GetDir( Drive : BYTE; VAR S : STRING );
BEGIN
  GetDir( Drive, S );
  CheckIO;
END;



PROCEDURE _MkDir( S : STRING );
BEGIN
  MkDir( S );
  CheckIO;
END;



PROCEDURE _Rename( VAR F; NewName : STRING );
BEGIN
  Rename( F, NewName );
  CheckIO;
END;



PROCEDURE _Reset( VAR F );
BEGIN
  Reset( F );
  CheckIO;
END;



PROCEDURE _Rewrite( VAR F );
BEGIN
  Rewrite( F );
  CheckIO;
END;



PROCEDURE _RmDir( S : STRING );
BEGIN
  RmDir( S );
  CheckIO;
END;



PROCEDURE _Seek( VAR F : FILE; N : LONGINT );
BEGIN
  Seek( F, N );
  CheckIO;
END;



PROCEDURE _Truncate( VAR F : FILE );
BEGIN
  Truncate( F );
  CheckIO;
END;




PROCEDURE SetString( VAR s : STRING; buffer : ARRAY OF CHAR; len : INTEGER );
BEGIN
  INLINE
  (
    int i = StringsLength( buffer );
    if (i < len)
    {
      len = i;
    }
    s.setLength(0);
    s.append( buffer );
    s.setLength( len );
  );
END;



FUNCTION StringOfChar( Ch : CHAR; Count : INTEGER ) : STRING;
BEGIN
  INLINE
  (
    java.lang.StringBuffer s = new java.lang.StringBuffer(Count);
    while (Count > 0)
    {
      s.append( Ch );
      Count -= 1;
    }
    Result = s;
  );
END;



FUNCTION StdinEoln : BOOLEAN;
CONST
  (*EOL = CHR(26);*)
  EOL = CHR(10);
VAR
  ch : CHAR;
BEGIN
  IF (termCH <> -1) THEN BEGIN
    ch := CHR(termCH);
  END
  ELSE BEGIN
    INLINE
    (
      try 
      {
        if (lookAhead == -1)
        {
          lookAhead = stdin.read();
        }
        ch = (char)lookAhead;
      }
      catch (java.io.IOException ex) 
      {
        ch = (char)0;
      }
    );
  END;
  Result := (ch = EOL);
  (*Result := Keypressed OR (termCH = ORD(EOL));*)
END;




FUNCTION TextEoln( VAR f : TEXT ) : BOOLEAN;
CONST
  EOL = CHR(10);
VAR
  EOLCH : CHAR;
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  EOLCH := EOL;
  INLINE
  (
    boolean ok = true;
    if (f == null)
    {
      f = stdtext;
    }
    if ((f.Handle == null) && (f.FileName == null))
    {
      Result = StdinEoln();
    }
    else if (f.Handle == null)
    {
      ok = false;
      Result = false;
    }
    else
    {
      try
      {
        long i = ((java.io.RandomAccessFile)f.Handle).getFilePointer();
        long j = ((java.io.RandomAccessFile)f.Handle).length();
        if (i < j)
        {
          char c = (char)((java.io.RandomAccessFile)f.Handle).read();
          ((java.io.RandomAccessFile)f.Handle).seek(i);
          Result = (c == EOLCH);
        }
        else
        {
          Result = true;
        }
      }
      catch (java.io.IOException e)
      {
        ok = false;
        Result = false;
      }
      if (!ok)
      {
        Result = false;
        InOutRes = 1;
      }
    }
  );
END;



FUNCTION _TextEoln( VAR f : TEXT ) : BOOLEAN;
BEGIN
  Result := TextEoln( f );
  CheckIO;
END;


FUNCTION _StdinEoln : BOOLEAN;
BEGIN
  Result := StdinEoln;
  CheckIO;
END;



PROCEDURE Flush( VAR f : TEXT );
BEGIN
  IF InOutRes <> 0 THEN BEGIN
    EXIT;
  END;
  INLINE
  (
    boolean ok = false;
    if (f == null)
    {
      f = stdtext;
    }
    if ((f.Handle == null) && (f.FileName == null))
    {
      try
      {
        stdout.flush();
        ok = true;
      }
      catch (java.io.IOException e)
      {
        ok = false;
      }
    }
    else if (f.Handle == null)
    {
      ok = false;
    }
    else
    {
      ok = true;
    }
    if (!ok)
    {
      InOutRes = 1;
    }
  );
END;



PROCEDURE _Flush( VAR  f : TEXT );
BEGIN
  Flush( f );
  CheckIO;
END;



PROCEDURE InitArray( arrayref : POINTER );
BEGIN
  (* Runtime function for allocating structured array components *)
  INLINE
  (
    int i,j;
    java.lang.Class ComponentType = arrayref.getClass().getComponentType();
    boolean is_array_component = ComponentType.isArray();
    java.lang.Object[] arrayobj = (java.lang.Object[])arrayref;
    j = arrayobj.length;
    for (i=0; i<j; i++)
    {
      if (is_array_component)
      {
        InitArray( arrayobj[i] );
      }
      else
      {
        try
        {
          arrayobj[i] = ComponentType.newInstance();
        }
        catch (java.lang.InstantiationException e)
        {
          throw new java.lang.RuntimeException
          ( "Array component not created because of InstantiationException" );
        }
        catch (java.lang.IllegalAccessException e)
        {
          throw new java.lang.RuntimeException
          ( "Array component not created because of IllegalAccessException" );
        }
      }
    }
  );
END;



FUNCTION CloneArray( arrayref : POINTER ) : POINTER;
BEGIN
  (* Runtime function for non-shallow array cloning,
     only used for compiler versions >= 3.0
  *)
  INLINE
  (

    int                      i,j;
    java.lang.Class[]        signature;
    java.lang.reflect.Method clone_method = null;
    java.lang.Class          ComponentType = arrayref.getClass().getComponentType();
    java.lang.Object[]       arrayobj;

    if (ComponentType.isArray())
    {
      arrayref = ((java.lang.Object[])arrayref).clone();
      arrayobj = (java.lang.Object[])arrayref;
      j = arrayobj.length;
      for (i=0; i<j; i++)
      {
        arrayobj[i] = CloneArray( arrayobj[i] );
      }
      Result = arrayobj;
    }
    else if (ComponentType == java.lang.Boolean.TYPE)
    {
      Result = ((boolean[])arrayref).clone();
    }
    else if (ComponentType == java.lang.Character.TYPE)
    {
      Result = ((char[])arrayref).clone();
    }
    else if (ComponentType == java.lang.Byte.TYPE)
    {
      Result = ((byte[])arrayref).clone();
    }
    else if (ComponentType == java.lang.Short.TYPE)
    {
      Result = ((short[])arrayref).clone();
    }
    else if (ComponentType == java.lang.Integer.TYPE)
    {
      Result = ((int[])arrayref).clone();
    }
    else if (ComponentType == java.lang.Long.TYPE)
    {
      Result = ((long[])arrayref).clone();
    }
    else if (ComponentType == java.lang.Float.TYPE)
    {
      Result = ((float[])arrayref).clone();
    }
    else if (ComponentType == java.lang.Double.TYPE)
    {
      Result = ((double[])arrayref).clone();
    }
    else if (ComponentType == java.lang.Void.TYPE)
    {
      Result = arrayref;
      SYSTEM.CANCEL
      (
        new java.lang.RuntimeException( "Invalid array of void" )
      );
    }
    else /* ComponentType is a class or interface */
    {
      signature = new java.lang.Class[1];
      signature[0] = java.lang.Void.TYPE;
      arrayref = ((java.lang.Object[])arrayref).clone();
      arrayobj = (java.lang.Object[])arrayref;
      try
      {
        clone_method = ComponentType.getMethod( "_Clone", signature );
        if (clone_method == null)
        {
          clone_method = ComponentType.getMethod( "Clone", signature );
        }
        if (clone_method != null)
        {
          j = arrayobj.length;
          for (i=0; i<j; i++)
          {
            arrayobj[i] = clone_method.invoke( arrayobj[i], null );
          }
        }
      }
      catch (java.lang.Exception e)
      {
        SYSTEM.CANCEL( e );
      }
      if (clone_method == null)
      {
        throw new java.lang.RuntimeException
        ( "Array component(s) are not cloneable" );
      }
      Result = arrayobj;
    }
  )
END;



PROCEDURE CopyArray( dest : POINTER; origin : POINTER );
BEGIN
  (* Runtime procedure for non-shallow array copy,
     only used for compiler versions >= 3.0
  *)
  INLINE
  (
    int                      i,j;
    java.lang.Class          ComponentType = dest.getClass().getComponentType();
    java.lang.Object[]       to;
    java.lang.Object[]       from;
    java.lang.Class[]        signature;
    java.lang.reflect.Method copy_method = null;
    java.lang.Object[]       args;

    if (ComponentType.isArray())
    {
      to = ((java.lang.Object[])dest);
      from = ((java.lang.Object[])origin);
      j = to.length;
      for (i=0; i<j; i++)
      {
        CopyArray( to[i], from[i] );
      }
    }
    else if (ComponentType == java.lang.Boolean.TYPE)
    {
      java.lang.System.arraycopy
      ( origin, 0, dest, 0, ((boolean[])dest).length );
    }
    else if (ComponentType == java.lang.Character.TYPE)
    {
      java.lang.System.arraycopy
      ( origin, 0, dest, 0, ((char[])dest).length );
    }
    else if (ComponentType == java.lang.Byte.TYPE)
    {
      java.lang.System.arraycopy
      ( origin, 0, dest, 0, ((byte[])dest).length );
    }
    else if (ComponentType == java.lang.Short.TYPE)
    {
      java.lang.System.arraycopy
      ( origin, 0, dest, 0, ((short[])dest).length );
    }
    else if (ComponentType == java.lang.Integer.TYPE)
    {
      java.lang.System.arraycopy
      ( origin, 0, dest, 0, ((int[])dest).length );
    }
    else if (ComponentType == java.lang.Long.TYPE)
    {
      java.lang.System.arraycopy
      ( origin, 0, dest, 0, ((long[])dest).length );
    }
    else if (ComponentType == java.lang.Float.TYPE)
    {
      java.lang.System.arraycopy
      ( origin, 0, dest, 0, ((float[])dest).length );
    }
    else if (ComponentType == java.lang.Double.TYPE)
    {
      java.lang.System.arraycopy
      ( origin, 0, dest, 0, ((double[])dest).length );
    }
    else if (ComponentType == java.lang.Void.TYPE)
    {
      SYSTEM.CANCEL
      (
        new java.lang.RuntimeException( "Invalid array of void" )
      );
    }
    else /* ComponentType is a class or interface */
    {
      signature = new java.lang.Class[1];
      signature[0] = ComponentType;
      to = (java.lang.Object[])dest;
      from = (java.lang.Object[])origin;
      java.lang.System.arraycopy( from, 0, to, 0, to.length );
      try
      {
        copy_method = ComponentType.getMethod( "_Copy", signature );
        if (copy_method != null)
        {
          j = to.length;
          for (i=0; i<j; i++)
          {
            args = new java.lang.Object[1];
            args[0] = from[i];
            copy_method.invoke( to[i], args );
          }
        }
      }
      catch (java.lang.Exception e)
      {
        SYSTEM.CANCEL( e );
      }
      if (copy_method == null)
      {
        throw new java.lang.RuntimeException
        ( "Array component(s) have no method _Copy" );
      }
    }
  )
END;


PROCEDURE AssignString( VAR dest : STRING; VAR source );
BEGIN
  { runtime function for Pascal STRING assignment }
  INLINE
  (
    if (source instanceof java.lang.String)
    {
      dest.setLength( 0 );
      dest.append( (java.lang.String)source );
    }
    else /* source instanceof java.lang.StringBuffer */
    {
      if (dest != (java.lang.StringBuffer)source)
      {
        dest.setLength(0);
        dest.append( ((java.lang.StringBuffer)source).toString() );
      }
    }
  );
END;




INITIALIZATION
  InOutRes := 0;
  NEW( stdtext );
  stdtext^.Handle := NIL;
  stdtext^.FileName := NIL;
  INLINE
  (
    args = _args;
    stdin = new java.io.InputStreamReader( java.lang.System.in );
    stdout = new java.io.OutputStreamWriter( java.lang.System.out );
  );
  vENotImplemented := NEW( ENotImplemented );
  TermCH := -1;
  lookAhead := -1;


END.

